# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'game_input_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_GameInputDialog(object):
    def setupUi(self, GameInputDialog):
        if not GameInputDialog.objectName():
            GameInputDialog.setObjectName(u"GameInputDialog")
        GameInputDialog.resize(508, 270)
        self.gridLayout = QGridLayout(GameInputDialog)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.accept_button = QPushButton(GameInputDialog)
        self.accept_button.setObjectName(u"accept_button")

        self.gridLayout.addWidget(self.accept_button, 9, 0, 1, 1)

        self.input_file_button = QPushButton(GameInputDialog)
        self.input_file_button.setObjectName(u"input_file_button")
        self.input_file_button.setMaximumSize(QSize(100, 16777215))

        self.gridLayout.addWidget(self.input_file_button, 2, 2, 1, 1)

        self.output_file_button = QPushButton(GameInputDialog)
        self.output_file_button.setObjectName(u"output_file_button")
        self.output_file_button.setMaximumSize(QSize(100, 16777215))

        self.gridLayout.addWidget(self.output_file_button, 5, 2, 1, 1)

        self.input_file_label = QLabel(GameInputDialog)
        self.input_file_label.setObjectName(u"input_file_label")
        self.input_file_label.setMaximumSize(QSize(16777215, 20))

        self.gridLayout.addWidget(self.input_file_label, 1, 0, 1, 2)

        self.input_file_edit = QLineEdit(GameInputDialog)
        self.input_file_edit.setObjectName(u"input_file_edit")

        self.gridLayout.addWidget(self.input_file_edit, 2, 0, 1, 2)

        self.output_format_layout = QHBoxLayout()
        self.output_format_layout.setSpacing(6)
        self.output_format_layout.setObjectName(u"output_format_layout")

        self.gridLayout.addLayout(self.output_format_layout, 7, 0, 1, 1)

        self.cancel_button = QPushButton(GameInputDialog)
        self.cancel_button.setObjectName(u"cancel_button")

        self.gridLayout.addWidget(self.cancel_button, 9, 2, 1, 1)

        self.description_label = QLabel(GameInputDialog)
        self.description_label.setObjectName(u"description_label")
        self.description_label.setWordWrap(True)

        self.gridLayout.addWidget(self.description_label, 0, 0, 1, 3)

        self.output_file_edit = QLineEdit(GameInputDialog)
        self.output_file_edit.setObjectName(u"output_file_edit")

        self.gridLayout.addWidget(self.output_file_edit, 5, 0, 1, 2)

        self.output_file_label = QLabel(GameInputDialog)
        self.output_file_label.setObjectName(u"output_file_label")
        self.output_file_label.setMaximumSize(QSize(16777215, 20))

        self.gridLayout.addWidget(self.output_file_label, 4, 0, 1, 2)

        self.auto_save_spoiler_check = QCheckBox(GameInputDialog)
        self.auto_save_spoiler_check.setObjectName(u"auto_save_spoiler_check")

        self.gridLayout.addWidget(self.auto_save_spoiler_check, 8, 0, 1, 1)


        self.retranslateUi(GameInputDialog)

        QMetaObject.connectSlotsByName(GameInputDialog)
    # setupUi

    def retranslateUi(self, GameInputDialog):
        GameInputDialog.setWindowTitle(QCoreApplication.translate("GameInputDialog", u"Game Patching", None))
        self.accept_button.setText(QCoreApplication.translate("GameInputDialog", u"Accept", None))
        self.input_file_button.setText(QCoreApplication.translate("GameInputDialog", u"Select File", None))
        self.output_file_button.setText(QCoreApplication.translate("GameInputDialog", u"Select File", None))
        self.input_file_label.setText(QCoreApplication.translate("GameInputDialog", u"Input File (Vanilla Gamecube ISO)", None))
        self.input_file_edit.setPlaceholderText(QCoreApplication.translate("GameInputDialog", u"Path to vanilla Gamecube ISO", None))
        self.cancel_button.setText(QCoreApplication.translate("GameInputDialog", u"Cancel", None))
        self.description_label.setText(QCoreApplication.translate("GameInputDialog", u"<html><head/><body><p>In order to create the randomized game, a ISO file of Metroid Prime 2: Echoes for the Nintendo Gamecube is necessary.</p><p>After using it once, a copy is kept by Randovania for later use.</p></body></html>", None))
        self.output_file_edit.setPlaceholderText(QCoreApplication.translate("GameInputDialog", u"Path where to place randomized game", None))
        self.output_file_label.setText(QCoreApplication.translate("GameInputDialog", u"Output File", None))
        self.auto_save_spoiler_check.setText(QCoreApplication.translate("GameInputDialog", u"Include a spoiler log on same directory", None))
    # retranslateUi




from unittest.mock import MagicMock

from PySide2.QtWidgets import QWidget

from randovania.games.game import RandovaniaGame
from randovania.gui.dialog import trick_usage_popup


def test_click_on_link(echoes_game_description,
                       skip_qtbot,
                       default_echoes_preset):
    # Setup
    main_window = QWidget()
    main_window.open_data_visualizer_at = MagicMock()
    skip_qtbot.add_widget(main_window)
    world_name = "World"
    area_name = "Area"

    # Run
    popup = trick_usage_popup.TrickUsagePopup(
        main_window,
        main_window,
        default_echoes_preset
    )
    popup._on_click_link_to_data_editor(f"data-editor://{world_name}/{area_name}")

    # Assert
    main_window.open_data_visualizer_at.assert_called_once_with(world_name, area_name,
                                                                game=RandovaniaGame.METROID_PRIME_ECHOES)
    assert popup.area_list_label.text() == ""



import dataclasses
from typing import FrozenSet

from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.layout.base.available_locations import RandomizationMode
from randovania.layout.base.logical_resource_action import LayoutLogicalResourceAction


@dataclasses.dataclass(frozen=True)
class FillerConfiguration:
    randomization_mode: RandomizationMode
    minimum_random_starting_items: int
    maximum_random_starting_items: int
    indices_to_exclude: FrozenSet[PickupIndex]
    multi_pickup_placement: bool
    logical_resource_action: LayoutLogicalResourceAction



import logging

_CURRENT_OPTIONS_FILE_VERSION = 16


def _convert_logic(layout_logic: str) -> str:
    if layout_logic == "no-glitches":
        return "no-tricks"
    return layout_logic


def _convert_v11(options: dict) -> dict:
    options.pop("layout_configuration", None)
    options.pop("patcher_configuration", None)

    return options


def _convert_v12(options: dict) -> dict:
    if "cosmetic_patches" in options:
        options["cosmetic_patches"]["unvisited_room_names"] = True

    return options


def _convert_v13(options: dict) -> dict:
    cosmetic_patches = options.pop("cosmetic_patches", None)
    output_directory = options.pop("output_directory", None)

    if cosmetic_patches is not None or output_directory is not None:
        if cosmetic_patches is None:
            cosmetic_patches = {
                "disable_hud_popup": True,
                "speed_up_credits": True,
                "open_map": True,
                "unvisited_room_names": True,
                "pickup_markers": True,
                "teleporter_sounds": True,
                "user_preferences": {
                    "sound_mode": 1,
                    "screen_brightness": 4,
                    "screen_x_offset": 0,
                    "screen_y_offset": 0,
                    "screen_stretch": 0,
                    "sfx_volume": 105,
                    "music_volume": 79,
                    "hud_alpha": 255,
                    "helmet_alpha": 255,
                    "hud_lag": False,
                    "invert_y_axis": False,
                    "rumble": True,
                    "hint_system": False
                }
            }
        options["per_game_options"] = {
            "prime2": {
                "cosmetic_patches": cosmetic_patches,
                "input_path": None,
                "output_directory": output_directory,
                "output_format": None,
            }
        }

    return options


def _convert_v14(options: dict) -> dict:
    for game_options in options.get("per_game_options", {}).values():
        game_options["output_format"] = None

    return options


def _convert_v15(options: dict) -> dict:
    per_game_options = options.get("per_game_options", {})
    if "prime1" in per_game_options and "cosmetic_patches" in per_game_options["prime1"]:
        per_game_options["prime1"]["cosmetic_patches"].pop("debug_pickups", None)

    return options


_CONVERTER_FOR_VERSION = {
    11: _convert_v11,
    12: _convert_v12,
    13: _convert_v13,
    14: _convert_v14,
    15: _convert_v15,
}


# debug_locations_check

def get_persisted_options_from_data(persisted_data: dict) -> dict:
    version = persisted_data.get("version", 0)
    options = persisted_data.get("options")

    if not isinstance(options, dict):
        logging.error("Data has no options.")
        return {}

    while version < _CURRENT_OPTIONS_FILE_VERSION:
        converter = _CONVERTER_FOR_VERSION.get(version)
        if converter is None:
            logging.error("Converter not found for version '{}'".format(version))
            return {}

        options = converter(options)
        version += 1

    if version > _CURRENT_OPTIONS_FILE_VERSION:
        logging.error("Options has an version from the future '{}'. Supported is only up to {}".format(
            version, _CURRENT_OPTIONS_FILE_VERSION))
        return {}

    return options


def serialized_data_for_options(data_to_persist: dict) -> dict:
    return {
        "version": _CURRENT_OPTIONS_FILE_VERSION,
        "options": data_to_persist
    }



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_hints.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetEchoesHints(object):
    def setupUi(self, PresetEchoesHints):
        if not PresetEchoesHints.objectName():
            PresetEchoesHints.setObjectName(u"PresetEchoesHints")
        PresetEchoesHints.resize(423, 259)
        self.centralWidget = QWidget(PresetEchoesHints)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.hint_layout = QVBoxLayout(self.centralWidget)
        self.hint_layout.setSpacing(6)
        self.hint_layout.setContentsMargins(11, 11, 11, 11)
        self.hint_layout.setObjectName(u"hint_layout")
        self.hint_layout.setContentsMargins(4, 8, 4, 0)
        self.hint_sky_temple_key_label = QLabel(self.centralWidget)
        self.hint_sky_temple_key_label.setObjectName(u"hint_sky_temple_key_label")
        self.hint_sky_temple_key_label.setWordWrap(True)

        self.hint_layout.addWidget(self.hint_sky_temple_key_label)

        self.hint_sky_temple_key_combo = QComboBox(self.centralWidget)
        self.hint_sky_temple_key_combo.addItem("")
        self.hint_sky_temple_key_combo.addItem("")
        self.hint_sky_temple_key_combo.addItem("")
        self.hint_sky_temple_key_combo.setObjectName(u"hint_sky_temple_key_combo")

        self.hint_layout.addWidget(self.hint_sky_temple_key_combo)

        PresetEchoesHints.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesHints)

        QMetaObject.connectSlotsByName(PresetEchoesHints)
    # setupUi

    def retranslateUi(self, PresetEchoesHints):
        PresetEchoesHints.setWindowTitle(QCoreApplication.translate("PresetEchoesHints", u"Hints", None))
        self.hint_sky_temple_key_label.setText(QCoreApplication.translate("PresetEchoesHints", u"<html><head/><body><p>This controls how precise the hints for Sky Temple Keys in Sky Temple Gateway are.</p><p><span style=\" font-weight:600;\">No hints</span>: The scans provide no useful information.</p><p><span style=\" font-weight:600;\">Show only the area name</span>: Each scan says the corresponding key is in &quot;Temple Grounds&quot;, &quot;Agon Wastes&quot;, etc.<br/>Aether and Dark Aether are distinguished; for example, &quot;Agon Wastes&quot; refers only to the light world.</p><p><span style=\" font-weight:600;\">Show area and room name</span>: Each scan says the corresponding key is in &quot;Great Temple - Transport A Access&quot;, &quot;Ing Hive - Hive Entrance&quot;, etc. For rooms with more than one item location, there's no way to distinguish which one of them that key is in.</p></body></html>", None))
        self.hint_sky_temple_key_combo.setItemText(0, QCoreApplication.translate("PresetEchoesHints", u"No hints", None))
        self.hint_sky_temple_key_combo.setItemText(1, QCoreApplication.translate("PresetEchoesHints", u"Show only the area name", None))
        self.hint_sky_temple_key_combo.setItemText(2, QCoreApplication.translate("PresetEchoesHints", u"Show area and room name", None))

    # retranslateUi




import dataclasses

from PySide2.QtWidgets import QWidget

from randovania.gui.dialog.base_cosmetic_patches_dialog import BaseCosmeticPatchesDialog
from randovania.gui.generated.prime_cosmetic_patches_dialog_ui import Ui_PrimeCosmeticPatchesDialog
from randovania.layout.prime1.prime_cosmetic_patches import PrimeCosmeticPatches


class PrimeCosmeticPatchesDialog(BaseCosmeticPatchesDialog, Ui_PrimeCosmeticPatchesDialog):
    _cosmetic_patches: PrimeCosmeticPatches

    def __init__(self, parent: QWidget, current: PrimeCosmeticPatches):
        super().__init__(parent)
        self.setupUi(self)
        self._cosmetic_patches = current

        self.on_new_cosmetic_patches(current)
        self.connect_signals()

    def connect_signals(self):
        super().connect_signals()

        self.qol_cosmetic_check.stateChanged.connect(self._persist_option_then_notify("qol_cosmetic"))
        self.open_map_check.stateChanged.connect(self._persist_option_then_notify("open_map"))

    def on_new_cosmetic_patches(self, patches: PrimeCosmeticPatches):
        self.qol_cosmetic_check.setChecked(patches.qol_cosmetic)
        self.open_map_check.setChecked(patches.open_map)

    def _persist_option_then_notify(self, attribute_name: str):
        def persist(value: int):
            self._cosmetic_patches = dataclasses.replace(
                self._cosmetic_patches,
                **{attribute_name: bool(value)}
            )

        return persist

    @property
    def cosmetic_patches(self) -> PrimeCosmeticPatches:
        return self._cosmetic_patches

    def reset(self):
        self.on_new_cosmetic_patches(PrimeCosmeticPatches())



from typing import TypeVar, List

from randovania.game_description.resources.resource_type import ResourceType


class MissingResource(ValueError):
    pass


T = TypeVar("T")


def find_resource_info_with_id(info_list: List[T], index: int, resource_type: ResourceType) -> T:
    for info in info_list:
        if info.index == index:
            return info
    indices = [info.index for info in info_list]
    raise MissingResource(f"{resource_type} Resource with index {index} not found in {indices}")


def find_resource_info_with_long_name(info_list: List[T], long_name: str) -> T:
    for info in info_list:
        if info.long_name == long_name:
            return info
    raise MissingResource(f"Resource with long_name '{long_name}' not found in {len(info_list)} resources")



from typing import List

from randovania.game_description.resources import resource_info
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import CurrentResources
from randovania.generator.item_pool import PoolResults, pickup_creator
from randovania.layout.prime1.artifact_mode import LayoutArtifactMode


def add_artifacts(resource_database: ResourceDatabase,
                  mode: LayoutArtifactMode,
                  artifact_minimum_progression: int,
                  ) -> PoolResults:
    """
    :param resource_database:
    :param mode
    :param artifact_minimum_progression
    :return:
    """
    item_pool: List[PickupEntry] = []
    initial_resources: CurrentResources = {}

    artifacts_to_place = mode.value

    for i in range(artifacts_to_place):
        item_pool.append(pickup_creator.create_artifact(i, artifact_minimum_progression, resource_database))

    first_automatic_artifact = artifacts_to_place

    for automatic_artifact in range(first_automatic_artifact, 12):
        resource_info.add_resource_gain_to_current_resources(
            pickup_creator.create_artifact(automatic_artifact, artifact_minimum_progression,
                                           resource_database).all_resources,
            initial_resources
        )

    return PoolResults(item_pool, {}, initial_resources)




from typing import List

from PySide2 import QtWidgets

from randovania.game_description import default_database
from randovania.game_description.item.item_database import ItemDatabase
from randovania.games.game import RandovaniaGame
from randovania.gui.preset_settings.metroid_item_pool_tab import MetroidPresetItemPool
from randovania.gui.preset_settings.split_ammo_widget import SplitAmmoWidget
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.preset import Preset


class EchoesPresetItemPool(MetroidPresetItemPool):
    _split_ammo_widgets: List[SplitAmmoWidget]

    def __init__(self, editor: PresetEditor):
        super().__init__(editor)
        item_database = default_database.item_database_for_game(self.game)

        self._create_split_ammo_widgets(item_database)

    def on_preset_changed(self, preset: Preset):
        super().on_preset_changed(preset)

        for split_ammo in self._split_ammo_widgets:
            split_ammo.on_preset_changed(preset, self._ammo_pickup_widgets)

    def _create_split_ammo_widgets(self, item_database: ItemDatabase):
        parent, layout, _ = self._boxes_for_category["beam"]

        self._split_ammo_widgets = []

        if self.game == RandovaniaGame.METROID_PRIME_ECHOES:
            beam_ammo = SplitAmmoWidget(
                parent, self._editor,
                unified_ammo=item_database.ammo["Beam Ammo Expansion"],
                split_ammo=[
                    item_database.ammo["Dark Ammo Expansion"],
                    item_database.ammo["Light Ammo Expansion"],
                ],
            )
            beam_ammo.setText("Split Beam Ammo Expansions")
            self._split_ammo_widgets.append(beam_ammo)

        if self._split_ammo_widgets:
            line = QtWidgets.QFrame(parent)
            line.setFrameShape(QtWidgets.QFrame.HLine)
            line.setFrameShadow(QtWidgets.QFrame.Sunken)
            layout.addWidget(line, layout.rowCount(), 0, 1, -1)

        for widget in self._split_ammo_widgets:
            layout.addWidget(widget, layout.rowCount(), 0, 1, -1)



from typing import Union

from randovania.game_description.world.node import ResourceNode
from randovania.generator.filler.pickup_list import PickupCombination

Action = Union[ResourceNode, PickupCombination]


def action_name(action: Action) -> str:
    if isinstance(action, tuple):
        return "[{}]".format(", ".join(a.name for a in action))
    else:
        return f"({type(action).__name__}) {action.name}"



from typing import Iterator, Dict

from randovania.game_description.game_description import GameDescription
from randovania.game_description.world.node import ResourceNode
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_info import ResourceInfo
from randovania.generator.filler.filler_library import UncollectedState, find_node_with_resource
from randovania.generator.generator_reach import GeneratorReach
from randovania.resolver import debug


def debug_print_collect_event(event: ResourceNode, game: GameDescription):
    if debug.debug_level() > 0:
        print("\n--> Collecting {}".format(game.world_list.node_name(event, with_world=True)))


def print_retcon_loop_start(game: GameDescription,
                            pickups_left: Iterator[PickupEntry],
                            reach: GeneratorReach,
                            player_index: int,
                            ):
    if debug.debug_level() > 0:
        current_uncollected = UncollectedState.from_reach(reach)
        if debug.debug_level() > 1:
            extra = ", pickups_left:\n{}".format(sorted(set(pickup.name for pickup in pickups_left)))
        else:
            extra = ""

        print("\n===============================")
        print("\n>>> Player {}: From {}, {} open pickup indices, {} open events{}".format(
            player_index,
            game.world_list.node_name(reach.state.node, with_world=True),
            len(current_uncollected.indices),
            len(current_uncollected.events),
            extra
        ))

        if debug.debug_level() > 2:
            print("\nCurrent reach:")
            for node in reach.nodes:
                print("[{!s:>5}, {!s:>5}] {}".format(reach.is_reachable_node(node), reach.is_safe_node(node),
                                                     game.world_list.node_name(node)))


def print_new_resources(game: GameDescription,
                        reach: GeneratorReach,
                        seen_count: Dict[ResourceInfo, int],
                        label: str,
                        ):
    world_list = game.world_list
    if debug.debug_level() > 1:
        for index, count in seen_count.items():
            if count == 1:
                node = find_node_with_resource(index, world_list.all_nodes)
                print("-> New {}: {}".format(label, world_list.node_name(node, with_world=True)))



from unittest.mock import MagicMock

from randovania.game_description.requirements import RequirementSet, RequirementList, ResourceRequirement
from randovania.game_description.resources import search
from randovania.generator.filler import pickup_list
from randovania.resolver import bootstrap


def test_requirement_lists_without_satisfied_resources(echoes_game_description):
    # Setup
    def item(name):
        return search.find_resource_info_with_long_name(echoes_game_description.resource_database.item, name)

    state = bootstrap.calculate_starting_state(echoes_game_description,
                                               echoes_game_description.create_game_patches(),
                                               100)
    state.resources[item("Seeker Launcher")] = 1
    state.resources[item("Space Jump Boots")] = 1
    uncollected_resources = []
    possible_sets = [
        RequirementSet([
            RequirementList([
                ResourceRequirement(item("Dark Visor"), 1, False),
                ResourceRequirement(item("Missile"), 5, False),
                ResourceRequirement(item("Seeker Launcher"), 1, False),
            ]),
            RequirementList([
                ResourceRequirement(item("Screw Attack"), 1, False),
                ResourceRequirement(item("Space Jump Boots"), 1, False),
            ]),
            RequirementList([
                ResourceRequirement(item("Power Bomb"), 1, False),
                ResourceRequirement(item("Boost Ball"), 1, False),
            ]),
        ]),
        RequirementSet([
            RequirementList([
                ResourceRequirement(item("Power Bomb"), 1, False),
                ResourceRequirement(item("Boost Ball"), 1, False),
            ]),
            RequirementList([
                ResourceRequirement(item("Spider Ball"), 1, False),
                ResourceRequirement(item("Boost Ball"), 1, False),
            ]),
        ]),
    ]

    # Run
    result = pickup_list._requirement_lists_without_satisfied_resources(
        state, possible_sets, uncollected_resources
    )

    # Assert
    assert result == {
        RequirementList([
            ResourceRequirement(item("Dark Visor"), 1, False),
            ResourceRequirement(item("Missile"), 5, False),
        ]),
        RequirementList([
            ResourceRequirement(item("Screw Attack"), 1, False),
        ]),
        RequirementList([
            ResourceRequirement(item("Power Bomb"), 1, False),
            ResourceRequirement(item("Boost Ball"), 1, False),
        ]),
        RequirementList([
            ResourceRequirement(item("Spider Ball"), 1, False),
            ResourceRequirement(item("Boost Ball"), 1, False),
        ]),
    }


def test_get_pickups_that_solves_unreachable(echoes_game_description, mocker):
    def item(name):
        return search.find_resource_info_with_long_name(echoes_game_description.resource_database.item, name)
    mock_req_lists: MagicMock = mocker.patch(
        "randovania.generator.filler.pickup_list._requirement_lists_without_satisfied_resources")

    pickups_left = []
    reach = MagicMock()
    reach.state.resources = {}
    reach.state.energy = 100
    possible_set = MagicMock()
    reach.unreachable_nodes_with_requirements.return_value = {"foo": possible_set}
    uncollected_resource_nodes = [MagicMock()]

    mock_req_lists.return_value = {
        RequirementList([
            ResourceRequirement(item("Dark Visor"), 1, False),
            ResourceRequirement(item("Missile"), 5, False),
        ]),
        RequirementList([
            ResourceRequirement(item("Screw Attack"), 1, False),
        ]),
        RequirementList([
            ResourceRequirement(item("Power Bomb"), 1, False),
            ResourceRequirement(item("Boost Ball"), 1, False),
        ]),
        RequirementList([
            ResourceRequirement(item("Spider Ball"), 1, False),
            ResourceRequirement(item("Boost Ball"), 1, False),
        ]),
    }

    # Run
    result = pickup_list.get_pickups_that_solves_unreachable(pickups_left, reach, uncollected_resource_nodes)

    # Assert
    mock_req_lists.assert_called_once_with(reach.state, [possible_set],
                                           [uncollected_resource_nodes[0].resource.return_value])
    assert result == tuple()




import re
from typing import Iterator, Dict

from PySide2 import QtWidgets
from PySide2.QtWidgets import QWidget

from randovania.game_description import default_database
from randovania.game_description.requirements import RequirementSet, ResourceRequirement
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import CurrentResources
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.game_description.world.area import Area
from randovania.game_description.world.node import DockNode
from randovania.gui.generated.trick_usage_popup_ui import Ui_TrickUsagePopup
from randovania.gui.lib.common_qt_lib import set_default_window_icon
from randovania.gui.lib.window_manager import WindowManager
from randovania.layout.base.trick_level import LayoutTrickLevel
from randovania.layout.base.trick_level_configuration import TrickLevelConfiguration
from randovania.layout.preset import Preset
from randovania.resolver import bootstrap


def _area_requirement_sets(area: Area,
                           database: ResourceDatabase,
                           ) -> Iterator[RequirementSet]:
    """
    Checks the area RequirementSet in the given Area uses the given trick at the given level.
    :param area:
    :param database:
    :return:
    """

    for node in area.nodes:
        if isinstance(node, DockNode):
            yield node.default_dock_weakness.requirement.as_set(database)

        for req in area.connections[node].values():
            yield req.as_set(database)


def _check_used_tricks(area: Area, trick_resources: CurrentResources, database: ResourceDatabase):
    result = set()

    for s in _area_requirement_sets(area, database):
        for alternative in s.alternatives:
            tricks: Dict[TrickResourceInfo, ResourceRequirement] = {
                req.resource: req
                for req in alternative.items
                if req.resource.resource_type == ResourceType.TRICK
            }
            if tricks and all(trick_resources[trick] >= tricks[trick].amount for trick in tricks):
                line = [
                    f"{trick.long_name} ({LayoutTrickLevel.from_number(req.amount).long_name})"
                    for trick, req in tricks.items()
                ]
                result.add(" and ".join(sorted(line)))

    return sorted(result)


class TrickUsagePopup(QtWidgets.QDialog, Ui_TrickUsagePopup):
    def __init__(self,
                 parent: QWidget,
                 window_manager: WindowManager,
                 preset: Preset,
                 ):

        super().__init__(parent)
        self.setupUi(self)
        set_default_window_icon(self)

        self._window_manager = window_manager
        self._game_description = default_database.game_description_for(preset.game)
        database = self._game_description.resource_database

        trick_level = preset.configuration.trick_level
        if trick_level.minimal_logic:
            trick_usage_description = "Minimal Logic"
        else:
            trick_usage_description = ", ".join(sorted(
                f"{trick.long_name} ({trick_level.level_for_trick(trick).long_name})"
                for trick in database.trick
                if trick_level.has_specific_level_for_trick(trick)
            ))

        # setup
        self.area_list_label.linkActivated.connect(self._on_click_link_to_data_editor)
        self.setWindowTitle("{} for preset {}".format(self.windowTitle(), preset.name))
        self.title_label.setText(self.title_label.text().format(
            trick_levels=trick_usage_description
        ))

        # connect
        self.button_box.accepted.connect(self.button_box_close)
        self.button_box.rejected.connect(self.button_box_close)

        if trick_level.minimal_logic:
            return

        # Update
        trick_resources = bootstrap.trick_resources_for_configuration(trick_level, database)

        lines = []

        for world in sorted(self._game_description.world_list.worlds, key=lambda it: it.name):
            for area in sorted(world.areas, key=lambda it: it.name):
                used_tricks = _check_used_tricks(area, trick_resources, database)
                if used_tricks:
                    lines.append(
                        f'<p><a href="data-editor://{world.correct_name(area.in_dark_aether)}/{area.name}">'
                        f'{world.correct_name(area.in_dark_aether)} - {area.name}</a>'
                        f'<br />{"<br />".join(used_tricks)}</p>'
                    )

        self.area_list_label.setText("".join(lines))

    def button_box_close(self):
        self.reject()

    def _on_click_link_to_data_editor(self, link: str):
        info = re.match(r"^data-editor://([^)]+)/([^)]+)$", link)
        if info:
            world_name, area_name = info.group(1, 2)
            self._window_manager.open_data_visualizer_at(world_name, area_name, game=self._game_description.game)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'scroll_label_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_ScrollLabelDialog(object):
    def setupUi(self, ScrollLabelDialog):
        if not ScrollLabelDialog.objectName():
            ScrollLabelDialog.setObjectName(u"ScrollLabelDialog")
        ScrollLabelDialog.resize(518, 382)
        self.root_layout = QVBoxLayout(ScrollLabelDialog)
        self.root_layout.setSpacing(6)
        self.root_layout.setContentsMargins(11, 11, 11, 11)
        self.root_layout.setObjectName(u"root_layout")
        self.root_layout.setContentsMargins(0, 0, 0, 0)
        self.scroll_area = QScrollArea(ScrollLabelDialog)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 516, 342))
        self.scroll_layout = QVBoxLayout(self.scroll_area_contents)
        self.scroll_layout.setSpacing(6)
        self.scroll_layout.setContentsMargins(11, 11, 11, 11)
        self.scroll_layout.setObjectName(u"scroll_layout")
        self.scroll_layout.setContentsMargins(2, 2, 2, 2)
        self.label = QLabel(self.scroll_area_contents)
        self.label.setObjectName(u"label")
        self.label.setMinimumSize(QSize(250, 0))
        self.label.setWordWrap(True)

        self.scroll_layout.addWidget(self.label)

        self.scroll_area.setWidget(self.scroll_area_contents)

        self.root_layout.addWidget(self.scroll_area)

        self.button_box_layout = QHBoxLayout()
        self.button_box_layout.setSpacing(6)
        self.button_box_layout.setObjectName(u"button_box_layout")
        self.button_box_layout.setContentsMargins(-1, -1, 6, 6)
        self.button_box = QDialogButtonBox(ScrollLabelDialog)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setMinimumSize(QSize(500, 0))
        self.button_box.setStandardButtons(QDialogButtonBox.Ok)

        self.button_box_layout.addWidget(self.button_box)


        self.root_layout.addLayout(self.button_box_layout)


        self.retranslateUi(ScrollLabelDialog)

        QMetaObject.connectSlotsByName(ScrollLabelDialog)
    # setupUi

    def retranslateUi(self, ScrollLabelDialog):
        ScrollLabelDialog.setWindowTitle(QCoreApplication.translate("ScrollLabelDialog", u"Message box", None))
        self.label.setText(QCoreApplication.translate("ScrollLabelDialog", u"<html><head/><body><p><span style=\" font-weight:600;\">TextLabel</span></p><p><br/></p></body></html>", None))
    # retranslateUi




from random import Random
from typing import Dict, List

from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import HintType, Hint, HintLocationPrecision
from randovania.game_description.world.world_list import WorldList
from randovania.games.prime.patcher_file_lib import hint_lib
from randovania.games.prime.patcher_file_lib.hint_formatters import LocationFormatter
from randovania.games.prime.patcher_file_lib.item_hints import create_pickup_hint
from randovania.games.prime.patcher_file_lib.temple_key_hint import create_temple_key_hint
from randovania.interface_common.players_configuration import PlayersConfiguration


class LocationHintCreator:
    world_list: WorldList
    area_namers: Dict[int, hint_lib.AreaNamer]
    joke_hints: List[str]

    def __init__(self, world_list: WorldList, area_namers: Dict[int, hint_lib.AreaNamer],
                 rng: Random, base_joke_hints: List[str]):
        self.world_list = world_list
        self.area_namers = area_namers
        self.rng = rng
        self.base_joke_hints = base_joke_hints
        self.joke_hints = []

    def create_joke_hint(self) -> str:
        if not self.joke_hints:
            self.joke_hints = sorted(self.base_joke_hints)
            self.rng.shuffle(self.joke_hints)
        return self.joke_hints.pop()

    def create_message_for_hint(self, hint: Hint,
                                all_patches: Dict[int, GamePatches],
                                players_config: PlayersConfiguration,
                                location_formatters: Dict[HintLocationPrecision, LocationFormatter],
                                ) -> str:
        if hint.hint_type == HintType.JOKE:
            return hint_lib.color_text(hint_lib.TextColor.JOKE, self.create_joke_hint())

        elif hint.hint_type == HintType.RED_TEMPLE_KEY_SET:
            return create_temple_key_hint(all_patches, players_config.player_index, hint.dark_temple, self.area_namers)

        else:
            assert hint.hint_type == HintType.LOCATION
            patches = all_patches[players_config.player_index]
            pickup_target = patches.pickup_assignment.get(hint.target)
            determiner, pickup_name = create_pickup_hint(patches.pickup_assignment,
                                                         self.world_list,
                                                         hint.precision.item,
                                                         pickup_target,
                                                         players_config,
                                                         hint.precision.include_owner)

            return location_formatters[hint.precision.location].format(
                determiner,
                hint_lib.color_text(hint_lib.TextColor.ITEM, pickup_name),
                hint,
            )



import copy
import dataclasses
from typing import Dict, Iterator, Tuple, List, TypeVar, Iterable

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder
from randovania.game_description import default_database
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.item.major_item import MajorItem
from randovania.games.game import RandovaniaGame
from randovania.layout.base.major_item_state import MajorItemState

T = TypeVar("T")


def _check_matching_items(actual: Iterable[T], reference: Iterable[T]):
    actual_items = set(actual)
    reference_items = set(reference)
    if actual_items != reference_items:
        raise ValueError(f"Non-matching keys: Has {actual_items}, expected {reference_items}.")


@dataclasses.dataclass(frozen=True)
class MajorItemsConfiguration(BitPackValue):
    game: RandovaniaGame
    items_state: Dict[MajorItem, MajorItemState]
    default_items: Dict[ItemCategory, MajorItem]
    minimum_random_starting_items: int
    maximum_random_starting_items: int

    def __post_init__(self):
        item_database = default_database.item_database_for_game(self.game)

        _check_matching_items(self.items_state.keys(), item_database.major_items.values())
        _check_matching_items(self.default_items.keys(), item_database.default_items.keys())

        for item, state in self.items_state.items():
            state.check_consistency(item)

        for category, options in item_database.default_items.items():
            if category not in self.default_items:
                raise ValueError(f"Category {category} is missing an item.")

            if self.default_items[category] not in options:
                raise ValueError(f"Category {category} has {self.default_items[category]} as default item, "
                                 f"but that's not a valid option.")

    @property
    def as_json(self) -> dict:
        return {
            "items_state": {
                major_item.name: state.as_json
                for major_item, state in self.items_state.items()
            },
            "default_items": {
                category.name: item.name
                for category, item in self.default_items.items()
            },
            "minimum_random_starting_items": self.minimum_random_starting_items,
            "maximum_random_starting_items": self.maximum_random_starting_items,
        }

    @classmethod
    def from_json(cls, value: dict, game: RandovaniaGame) -> "MajorItemsConfiguration":
        item_database = default_database.item_database_for_game(game)

        items_state = {}
        for name, item in item_database.major_items.items():
            if name in value["items_state"]:
                state = MajorItemState.from_json(value["items_state"][name])
            else:
                state = MajorItemState()
            items_state[item] = state

        default_items = {
            category: item_database.major_items[value["default_items"][category.name]]
            for category, _ in item_database.default_items.items()
        }

        return cls(
            game=game,
            items_state=items_state,
            default_items=default_items,
            minimum_random_starting_items=value["minimum_random_starting_items"],
            maximum_random_starting_items=value["maximum_random_starting_items"],
        )

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        reference: MajorItemsConfiguration = metadata["reference"]

        name_to_item: Dict[str, MajorItem] = {item.name: item for item in self.items_state.keys()}

        modified_items = sorted(
            item.name for item, state in self.items_state.items()
            if state != reference.items_state[item]
        )
        yield from bitpacking.pack_sorted_array_elements(modified_items, sorted(name_to_item.keys()))
        for item_name in modified_items:
            item = name_to_item[item_name]
            yield from self.items_state[item].bit_pack_encode(item, reference=reference.items_state[item])

        # default_items
        for category, default in self.default_items.items():
            all_major = [major for major in reference.items_state.keys() if major.item_category == category]
            yield from bitpacking.pack_array_element(default, all_major)

        # random starting items
        yield from bitpacking.encode_big_int(self.minimum_random_starting_items)
        yield from bitpacking.encode_big_int(self.maximum_random_starting_items)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "MajorItemsConfiguration":
        reference: MajorItemsConfiguration = metadata["reference"]

        name_to_item: Dict[str, MajorItem] = {item.name: item for item in reference.items_state.keys()}
        modified_items = bitpacking.decode_sorted_array_elements(decoder, sorted(name_to_item.keys()))

        items_state = copy.copy(reference.items_state)
        for item_name in modified_items:
            item = name_to_item[item_name]
            items_state[item] = MajorItemState.bit_pack_unpack(decoder, item,
                                                               reference=reference.items_state[item])

        # default_items
        default_items = {}
        for category in reference.default_items.keys():
            all_major = [major for major in reference.items_state.keys() if major.item_category == category]
            default_items[category] = decoder.decode_element(all_major)

        # random starting items
        minimum = bitpacking.decode_big_int(decoder)
        maximum = bitpacking.decode_big_int(decoder)

        return cls(
            game=reference.game,
            items_state=items_state,
            default_items=default_items,
            minimum_random_starting_items=minimum,
            maximum_random_starting_items=maximum,
        )

    def replace_state_for_item(self, item: MajorItem, state: MajorItemState) -> "MajorItemsConfiguration":
        return self.replace_states({
            item: state
        })

    def replace_states(self, new_states: Dict[MajorItem, MajorItemState]) -> "MajorItemsConfiguration":
        """
        Creates a copy of this MajorItemsConfiguration where the state of all given items are replaced by the given
        states.
        :param new_states:
        :return:
        """
        items_state = copy.copy(self.items_state)

        for item, state in new_states.items():
            items_state[item] = state

        return dataclasses.replace(self, items_state=items_state)

    def replace_default_item(self, category: ItemCategory, item: MajorItem) -> "MajorItemsConfiguration":
        """
        Creates a copy of this MajorItemsConfiguration where the default item for the given category
        is replaced by the given item.
        :param category:
        :param item:
        :return:
        """
        default_items = copy.copy(self.default_items)
        default_items[category] = item
        return dataclasses.replace(self, default_items=default_items)

    def calculate_provided_ammo(self) -> Dict[int, int]:
        result: Dict[int, int] = {}

        for item, state in self.items_state.items():
            total_pickups = state.num_shuffled_pickups + state.num_included_in_starting_items
            if state.include_copy_in_original_location:
                total_pickups += 1

            for ammo_index, ammo_count in zip(item.ammo_index, state.included_ammo):
                result[ammo_index] = result.get(ammo_index, 0) + ammo_count * total_pickups

        return result

    def dangerous_settings(self) -> List[str]:
        result = []
        for major_item, state in self.items_state.items():
            if (major_item.warning is not None and "EXPERIMENTAL" in major_item.warning
                    and state.num_included_in_starting_items == 0):
                result.append(f"Not starting with {major_item.name}")

        return result



import asyncio
import json
import logging
import os
import typing
import uuid
from pathlib import Path
from typing import List, Optional, Iterator, Dict

import dulwich.porcelain
import dulwich.repo

import randovania
from randovania.games.game import RandovaniaGame
from randovania.layout.preset_migration import VersionedPreset, InvalidPreset


def read_preset_list() -> List[Path]:
    base_path = randovania.get_data_path().joinpath("presets")

    with base_path.joinpath("presets.json").open() as presets_file:
        preset_list = json.load(presets_file)["presets"]

    return [
        base_path.joinpath(preset["path"])
        for preset in preset_list
    ]


def _commit(message: str, file_path: Path, repository: Path, remove: bool):

    with dulwich.porcelain.open_repo_closing(repository) as r:
        r = typing.cast(dulwich.repo.Repo, r)
        # Detect invalid index
        try:
            r.open_index()
        except Exception:
            os.remove(r.index_path())
            r.reset_index()

    author = "randovania <nobody@example.com>"
    if remove:
        dulwich.porcelain.remove(repository, [file_path])
    else:
        dulwich.porcelain.add(repository, [file_path])

    dulwich.porcelain.commit(repository, message=f"{message} using Randovania v{randovania.VERSION}",
                             author=author, committer=author)


class PresetManager:
    included_presets: Dict[uuid.UUID, VersionedPreset]
    custom_presets: Dict[uuid.UUID, VersionedPreset]
    _data_dir: Optional[Path]
    _fallback_dir: Optional[Path]

    def __init__(self, data_dir: Optional[Path]):
        self.logger = logging.getLogger("PresetManager")
        self.included_presets = {
            preset.uuid: preset
            for preset in [VersionedPreset.from_file_sync(f) for f in read_preset_list()]
        }

        self.custom_presets = {}
        if data_dir is not None:
            self._data_dir = data_dir
        else:
            self._data_dir = None

    async def load_user_presets(self):
        all_files = self._data_dir.glob(f"*.{VersionedPreset.file_extension()}")
        user_presets = await asyncio.gather(*[VersionedPreset.from_file(f) for f in all_files])
        for preset in typing.cast(List[VersionedPreset], user_presets):
            if preset.is_for_known_game():
                self.custom_presets[preset.uuid] = preset

    @property
    def default_preset(self) -> VersionedPreset:
        for preset in self.included_presets.values():
            return preset
        raise ValueError(f"No included presets")

    def default_preset_for_game(self, game: RandovaniaGame) -> VersionedPreset:
        for preset in self.included_presets.values():
            if preset.game == game:
                return preset
        raise ValueError(f"{game} has no included preset")

    def presets_for_game(self, game: RandovaniaGame) -> Iterator[VersionedPreset]:
        for preset in self.all_presets:
            if preset.game == game:
                yield preset

    @property
    def all_presets(self) -> Iterator[VersionedPreset]:
        yield from self.included_presets.values()
        yield from self.custom_presets.values()

    def _commit(self, message: str, file_path: Path, remove: bool):
        repo_root = self._data_dir.parent
        try:
            self.logger.info("Will perform git operation: %s for path %s", message, str(file_path))
            _commit(message, file_path, repo_root, remove)
            self.logger.debug("Git operation successful.")
        except Exception as e:
            self.logger.warning(f"Error committing change to presets: {e}", exc_info=e)

    def add_new_preset(self, new_preset: VersionedPreset) -> bool:
        """
        Adds a new custom preset.
        :param: new_preset
        :return True, if there wasn't any preset with that name
        """
        assert new_preset.uuid not in self.included_presets
        existed_before = new_preset.uuid in self.custom_presets
        self.custom_presets[new_preset.uuid] = new_preset

        path = self._file_name_for_preset(new_preset)
        new_preset.save_to_file(path)
        self._commit(f"Update preset '{new_preset.name}'", path, False)

        return not existed_before

    def delete_preset(self, preset: VersionedPreset):
        del self.custom_presets[preset.uuid]
        path = self._file_name_for_preset(preset)
        os.remove(path)
        self._commit(f"Remove preset '{preset.name}'", path, True)

    def included_preset_with(self, game: RandovaniaGame, name: str) -> Optional[VersionedPreset]:
        for preset in self.included_presets.values():
            if preset.game == game and preset.name == name:
                return preset

        return None

    def preset_for_uuid(self, the_uid: uuid.UUID) -> Optional[VersionedPreset]:
        return self.included_presets.get(the_uid, self.custom_presets.get(the_uid))

    def _file_name_for_preset(self, preset: VersionedPreset) -> Path:
        return self._data_dir.joinpath("{}.{}".format(preset.uuid, preset.file_extension()))

    def should_do_migration(self):
        if not self.custom_presets:
            from randovania.interface_common import persistence
            for _ in persistence.local_data_dir().joinpath("presets").glob("*.rdvpreset"):
                return True
        return False

    async def migrate_from_old_path(self, on_update):
        from randovania.interface_common import persistence

        files_to_commit = []
        self.logger.info("Performing migration of presets from old path")

        all_files = list(persistence.local_data_dir().joinpath("presets").glob("*.rdvpreset"))

        for i, old_file in enumerate(all_files):
            on_update(i, len(all_files))
            preset = await VersionedPreset.from_file(old_file)
            try:
                preset.ensure_converted()
                path = self._file_name_for_preset(preset)
                preset.save_to_file(path)
                self.custom_presets[preset.uuid] = preset
                files_to_commit.append(path)
                self.logger.info("Migrated %s", preset.name)

            except InvalidPreset as e:
                self.logger.warning(f"Not migrating {preset.name}: {e}")
                continue

        on_update(len(all_files), len(all_files))
        dulwich.porcelain.add(self._data_dir.parent, files_to_commit)
        self.logger.info("Finished migration from old path.")

        author = "randovania <nobody@example.com>"
        dulwich.porcelain.commit(self._data_dir.parent,
                                 message=f"Migrated old presets using Randovania v{randovania.VERSION}",
                                 author=author, committer=author)



from typing import List

from randovania.dol_patching.assembler import ppc


def _b(instruction: ppc.BaseInstruction, address: int = 0, symbols = None) -> List[int]:
    if symbols is None:
        symbols = {}
    return list(instruction.bytes_for(address, symbols=symbols))


def test_instruction_iter():
    assert _b(ppc.Instruction(1234)) == [0x00, 0x00, 0x04, 0xD2]


def test_instruction_eq():
    assert ppc.Instruction(1234) == ppc.Instruction(1234)


def test_instruction_with_label():
    instruction = ppc.Instruction(1234)
    assert instruction.label is None

    with_label = instruction.with_label("foobar")
    assert instruction is with_label
    assert instruction.label == "foobar"


def test_stw():
    assert _b(ppc.stw(ppc.r31, 0x1c, ppc.r1)) == [0x93, 0xe1, 0x00, 0x1c]


def test_stfs():
    assert _b(ppc.stfs(ppc.f0, 0x1000, ppc.r2)) == [208, 2, 16, 0]


def test_lfs():
    assert _b(ppc.lfs(ppc.f0, (0x8041a4a8 - 0x804223c0), ppc.r2)) == [0xc0, 0x02, 0x80, 0xe8]


def test_addi():
    assert _b(ppc.addi(ppc.r3, ppc.r1, 0x8)) == [0x38, 0x61, 0x00, 0x08]


def test_or():
    assert _b(ppc.or_(ppc.r31, ppc.r3, ppc.r3)) == [124, 127, 27, 120]


def test_lmw():
    assert _b(ppc.lmw(ppc.r25, 0x774, ppc.r25)) == [187, 57, 7, 116]


def test_lwz():
    assert _b(ppc.lwz(ppc.r10, 0x774, ppc.r25)) == [0x81, 0x59, 0x07, 0x74]


def test_lwzx():
    assert _b(ppc.lwzx(ppc.r0, ppc.r4, ppc.r0)) == [0x7c, 0x04, 0x00, 0x2e]


def test_lhz():
    assert _b(ppc.lhz(ppc.r6, -0x40da, ppc.r2)) == [0xa0, 0xc2, 0xbf, 0x26]


def test_lbz():
    assert _b(ppc.lbz(ppc.r4, 0x2, ppc.r3)) == [0x88, 0x83, 0x00, 0x02]


def test_rlwinm():
    assert _b(ppc.rlwinm(ppc.r0, ppc.r0, 0x2, 0x0, 0x1d)) == [0x54, 0x00, 0x10, 0x3a]


def test_cmpwi_a():
    assert _b(ppc.cmpwi(ppc.r30, 0)) == [0x2c, 0x1e, 0x00, 0x00]


def test_cmpwi_b():
    assert _b(ppc.cmpwi(ppc.r4, -1)) == [0x2c, 0x04, 0xFF, 0xFF]


def test_cmpw():
    assert _b(ppc.cmpwi(ppc.r4, -1)) == [0x2c, 0x04, 0xFF, 0xFF]


def test_b():
    assert _b(ppc.b(0x80085760), address=0x80085760 + 0x9C) == [0x4b, 0xff, 0xff, 0x64]


def test_bl():
    assert _b(ppc.bl(0x80085760), address=0x80085760 + 0x9C) == [0x4b, 0xff, 0xff, 0x65]


def test_bdnz():
    assert _b(ppc.bdnz(16, relative=True)) == [0x42, 0x00, 0x00, 0x10]


def test_beq():
    assert _b(ppc.beq(0x80038094), address=0x80038034) == [0x41, 0x82, 0x00, 0x60]


def test_beq_symbol():
    symbols = {"TheFunPlace": 0x80038094}
    assert _b(ppc.beq("TheFunPlace"), address=0x80038034, symbols=symbols) == [0x41, 0x82, 0x00, 0x60]


def test_bge():
    assert _b(ppc.bge(0x801c93f8), address=0x801c93ec) == [0x40, 0x80, 0x00, 0x0c]


def test_bne():
    assert _b(ppc.bne(0x80038094), address=0x80038034) == [0x40, 0x82, 0x00, 0x60]


def test_blr():
    assert _b(ppc.blr()) == [0x4e, 0x80, 0x00, 0x20]


def test_bctrl():
    assert _b(ppc.bctrl()) == [0x4E, 0x80, 0x04, 0x21]


def test_li():
    assert _b(ppc.li(ppc.r5, 9999)) == [56, 160, 39, 15]


def test_stwu():
    assert _b(ppc.stwu(ppc.r1, -0x2C, ppc.r1)) == [0x94, 0x21, 0xff, 0xd4]


def test_stmw():
    assert _b(ppc.stmw(ppc.r25, -0x2C, ppc.r1)) == [191, 33, 255, 212]


def test_sync():
    assert _b(ppc.sync()) == [0x7c, 0x00, 0x04, 0xac]


def test_isync():
    assert _b(ppc.isync()) == [0x4c, 0x00, 0x01, 0x2c]


def test_dcbi():
    assert _b(ppc.dcbi(1, 2)) == [0x7c, 0x01, 0x13, 0xac]


def test_mfspr_LR():
    assert _b(ppc.mfspr(ppc.r0, ppc.LR)) == [0x7c, 0x08, 0x02, 0xa6]


def test_mtctr():
    assert _b(ppc.mtctr(ppc.r12)) == [0x7d, 0x89, 0x03, 0xa6]



import struct
from typing import Optional, List, Tuple

from randovania.game_connection.connection_base import Inventory
from randovania.game_connection.connector.prime_remote_connector import PrimeRemoteConnector, DolRemotePatch
from randovania.game_connection.executor.memory_operation import MemoryOperation, MemoryOperationExecutor
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.world.world import World
from randovania.games.prime import (all_prime_dol_patches)
from randovania.games.prime.corruption_dol_patches import CorruptionDolVersion


def format_received_item(item_name: str, player_name: str) -> str:
    special = {
        "Locked Missile Expansion": ("Received Missile Expansion from {provider_name}, "
                                     "but the Missile Launcher is required to use it."),
        "Locked Ship Missile Expansion": ("Received Ship Missile Expansion from {provider_name}, "
                                          "but the main launcher is required to use it."),
    }

    generic = "Received {item_name} from {provider_name}."

    return special.get(item_name, generic).format(item_name=item_name, provider_name=player_name)


def _corruption_powerup_offset(item_index: int) -> int:
    powerups_offset = 0x50
    vector_data_offset = 0x4
    powerup_size = 0xc
    return (powerups_offset + vector_data_offset) + (item_index * powerup_size)


class CorruptionRemoteConnector(PrimeRemoteConnector):
    version: CorruptionDolVersion

    def __init__(self, version: CorruptionDolVersion):
        super().__init__(version)

    def _asset_id_format(self):
        return ">Q"

    async def current_game_status(self, executor: MemoryOperationExecutor) -> Tuple[bool, Optional[World]]:
        """
        Fetches the world the player's currently at, or None if they're not in-game.
        :param executor:
        :return: bool indicating if there's a pending `execute_remote_patches` operation.
        """

        cstate_manager_global = self.version.cstate_manager_global

        mlvl_offset = 8
        asset_id_size = struct.calcsize(self._asset_id_format())

        # TODO: there's one extra pointer indirection
        cplayer_offset = 40
        player_offset = 0x2184

        memory_ops = [
            MemoryOperation(self.version.game_state_pointer, offset=mlvl_offset, read_byte_count=asset_id_size),
            MemoryOperation(cstate_manager_global + 0x2, read_byte_count=1),
            MemoryOperation(cstate_manager_global + cplayer_offset, offset=player_offset, read_byte_count=4),
        ]
        results = await executor.perform_memory_operations(memory_ops)
        player_pointer = results.get(memory_ops[2])
        player_vtable = None
        if player_pointer is not None:
            player_vtable = await executor.perform_single_memory_operation(MemoryOperation(
                struct.unpack(">I", player_pointer)[0], read_byte_count=4,
            ))

        pending_op_byte = results[memory_ops[1]]
        has_pending_op = pending_op_byte != b"\x00"
        return has_pending_op, self._current_status_world(results.get(memory_ops[0]), player_vtable)

    async def _memory_op_for_items(self, executor: MemoryOperationExecutor, items: List[ItemResourceInfo],
                                   ) -> List[MemoryOperation]:
        player_state_pointer = int.from_bytes(await executor.perform_single_memory_operation(MemoryOperation(
            address=self.version.game_state_pointer,
            read_byte_count=4,
        )), "big") + 36

        return [
            MemoryOperation(
                address=player_state_pointer,
                offset=_corruption_powerup_offset(item.index),
                read_byte_count=8,
            )
            for item in items
        ]

    async def _patches_for_pickup(self, provider_name: str, pickup: PickupEntry, inventory: Inventory):
        item_name, resources_to_give = self._resources_to_give_for_pickup(pickup, inventory)

        self.logger.debug(f"Resource changes for {pickup.name} from {provider_name}: {resources_to_give}")
        patches = [
            all_prime_dol_patches.adjust_item_amount_and_capacity_patch(
                self.version.powerup_functions, self.game.game,
                item.index, delta,
            )
            for item, delta in resources_to_give.items()
        ]
        return patches, format_received_item(item_name, provider_name)

    def _dol_patch_for_hud_message(self, message: str) -> DolRemotePatch:
        raise RuntimeError("Unable to prepare dol patches for hud display in Corruption")

    async def find_missing_remote_pickups(self, executor: MemoryOperationExecutor, inventory: Inventory,
                                          remote_pickups: Tuple[Tuple[str, PickupEntry], ...],
                                          in_cooldown: bool,
                                          ) -> Tuple[List[DolRemotePatch], bool]:
        # Not yet implemented
        return [], False

    async def execute_remote_patches(self, executor: MemoryOperationExecutor, patches: List[DolRemotePatch]) -> None:
        raise RuntimeError("Unable to execute remote patches in Corruption")



import math
from typing import Iterable, Dict, NamedTuple, Optional

from PySide2 import QtWidgets
from PySide2.QtCore import Qt

from randovania.game_description.item.ammo import Ammo
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.base.ammo_state import AmmoState
from randovania.layout.preset import Preset


class AmmoPickupWidgets(NamedTuple):
    pickup_spinbox: QtWidgets.QSpinBox
    expected_count: QtWidgets.QLabel
    pickup_box: QtWidgets.QGroupBox
    require_major_item_check: Optional[QtWidgets.QCheckBox]


class SplitAmmoWidget(QtWidgets.QCheckBox):
    def __init__(self, parent: QtWidgets.QWidget, editor: PresetEditor,
                 unified_ammo: Ammo, split_ammo: Iterable[Ammo]):
        super().__init__(parent)
        self._editor = editor
        self.unified_ammo = unified_ammo
        self.split_ammo = list(split_ammo)
        self.setTristate(True)
        self._last_check_state = (True, False)
        self.clicked.connect(self.change_split)

        invalid_split = [split for split in self.split_ammo if len(split.items) != 1]
        if invalid_split:
            raise ValueError(f"The following split ammo have more than one item associated: {invalid_split}")

        if len(unified_ammo.items) != len(self.split_ammo):
            raise ValueError("The unified ammo should have as many items as there are split ammo items.")

    def on_preset_changed(self, preset: Preset, ammo_pickup_widgets: Dict[Ammo, AmmoPickupWidgets]):
        ammo_configuration = preset.configuration.ammo_configuration

        has_unified = ammo_configuration.items_state[self.unified_ammo].pickup_count > 0
        has_split = any(ammo_configuration.items_state[item].pickup_count > 0
                        for item in self.split_ammo)

        if not has_split and not has_unified:
            has_split, has_unified = self._last_check_state
        else:
            self._last_check_state = has_split, has_unified

        if has_unified:
            new_state = Qt.PartiallyChecked if has_split else Qt.Unchecked
        else:
            new_state = Qt.Checked
        self.setCheckState(new_state)

        ammo_pickup_widgets[self.unified_ammo].pickup_box.setVisible(has_unified)
        for item in self.split_ammo:
            ammo_pickup_widgets[item].pickup_box.setVisible(has_split)

    def change_split(self, has_split: bool):
        with self._editor as editor:
            ammo_configuration = editor.ammo_configuration

            current_total = sum(
                ammo_configuration.items_state[ammo].pickup_count
                for ammo in (self.unified_ammo, *self.split_ammo)
            )

            new_states = {}
            if has_split:
                ref = ammo_configuration.items_state[self.unified_ammo]
                for i, split in enumerate(self.split_ammo):
                    new_count = current_total // len(self.split_ammo)
                    new_states[split] = AmmoState(
                        ammo_count=(math.ceil(
                            ref.ammo_count[i] * (current_total / new_count),
                        ),),
                        pickup_count=new_count,
                    )
                new_states[self.unified_ammo] = AmmoState(
                    ammo_count=tuple(0 for _ in self.unified_ammo.items),
                    pickup_count=0,
                )

            else:
                for split in self.split_ammo:
                    new_states[split] = AmmoState(
                        ammo_count=(0,),
                        pickup_count=0
                    )
                new_states[self.unified_ammo] = AmmoState(
                    ammo_count=tuple(
                        math.ceil(
                            ammo_configuration.items_state[split].ammo_count[0] * (
                                    ammo_configuration.items_state[split].pickup_count / current_total
                            )
                        )
                        for split in self.split_ammo
                    ),
                    pickup_count=current_total,
                )

            editor.ammo_configuration = ammo_configuration.replace_states(new_states)
            self._last_check_state = (has_split, not has_split)



import os
import subprocess
from pathlib import Path
from typing import Union

import pytest
from PySide2.QtWidgets import QDialog
from mock import AsyncMock, MagicMock, ANY

from randovania.gui.main_window import MainWindow
from randovania.interface_common.options import Options
from randovania.interface_common.preset_manager import PresetManager


def create_window(options: Union[Options, MagicMock],
                  preset_manager: PresetManager) -> MainWindow:
    return MainWindow(options, preset_manager, MagicMock(), False)


@pytest.fixture(name="default_main_window")
def _default_main_window(skip_qtbot, preset_manager) -> MainWindow:
    return create_window(Options(MagicMock()), preset_manager)


def test_drop_random_event(default_main_window: MainWindow,
                           ):
    # Creating a window should not fail
    pass


@pytest.mark.parametrize(["url", "should_accept"], [
    ("something/game.iso", False),
    ("other/game.rdvgame", True),
    ("boss/custom.rdvpreset", True),
])
def test_dragEnterEvent(default_main_window: MainWindow, url, should_accept):
    mock_url = MagicMock()
    mock_url.toLocalFile.return_value = url
    event = MagicMock()
    event.mimeData.return_value.urls.return_value = [mock_url]

    # Run
    default_main_window.dragEnterEvent(event)

    # Assert
    if should_accept:
        event.acceptProposedAction.assert_called_once_with()
    else:
        event.acceptProposedAction.assert_not_called()


def test_drop_event_layout(default_main_window, mocker):
    mock_url = MagicMock()
    mock_url.toLocalFile.return_value = "/my/path.rdvgame"

    event = MagicMock()
    event.mimeData.return_value.urls.return_value = [mock_url]
    mock_from_file: MagicMock = mocker.patch("randovania.layout.layout_description.LayoutDescription.from_file")

    default_main_window.open_game_details = MagicMock()

    # Run
    default_main_window.dropEvent(event)

    # Assert
    mock_from_file.assert_called_once_with(Path("/my/path.rdvgame"))
    default_main_window.open_game_details.assert_called_once_with(mock_from_file.return_value)


@pytest.mark.asyncio
async def test_drop_event_preset(default_main_window):
    await default_main_window._initialize_post_show_body()

    mock_url = MagicMock()
    mock_url.toLocalFile.return_value = "/my/path.rdvpreset"
    event = MagicMock()
    event.mimeData.return_value.urls.return_value = [mock_url]
    default_main_window.generate_seed_tab.import_preset_file = MagicMock()

    # Run
    default_main_window.dropEvent(event)

    # Assert
    default_main_window.generate_seed_tab.import_preset_file(Path("/my/path.rdvpreset"))
    assert default_main_window.main_tab_widget.currentWidget() == default_main_window.tab_create_seed


@pytest.mark.asyncio
async def test_browse_racetime(default_main_window, mocker):
    mock_new_dialog = mocker.patch("randovania.gui.dialog.racetime_browser_dialog.RacetimeBrowserDialog")
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock,
                                       return_value=QDialog.Accepted)
    dialog = mock_new_dialog.return_value
    dialog.refresh = AsyncMock(return_value=True)
    default_main_window.generate_seed_from_permalink = AsyncMock()

    # Run
    await default_main_window._browse_racetime()

    # Assert
    mock_new_dialog.assert_called_once_with()
    dialog.refresh.assert_awaited_once_with()
    mock_execute_dialog.assert_awaited_once_with(dialog)
    default_main_window.generate_seed_from_permalink.assert_awaited_once_with(dialog.permalink)


@pytest.mark.asyncio
async def test_generate_seed_from_permalink(default_main_window, mocker):
    permalink = MagicMock()
    mock_generate_layout: MagicMock = mocker.patch("randovania.interface_common.simplified_patcher.generate_layout",
                                                   autospec=True)
    default_main_window.open_game_details = MagicMock()

    # Run
    await default_main_window.generate_seed_from_permalink(permalink)

    # Assert
    mock_generate_layout.assert_called_once_with(progress_update=ANY,
                                                 permalink=permalink,
                                                 options=default_main_window._options)
    default_main_window.open_game_details.assert_called_once_with(mock_generate_layout.return_value)


@pytest.mark.parametrize("os_type", ["Windows", "Darwin", "Linux"])
@pytest.mark.parametrize("throw_exception", [True, False])
def test_on_menu_action_previously_generated_games(default_main_window, mocker, os_type, throw_exception, monkeypatch):
    mock_start_file = MagicMock()
    mock_subprocess_run = MagicMock()
    monkeypatch.setattr(os, "startfile", mock_start_file, raising=False)
    monkeypatch.setattr(subprocess, "run", mock_subprocess_run, raising=False)
    mocker.patch("platform.system", return_value=os_type)
    mock_message_box = mocker.patch("PySide2.QtWidgets.QMessageBox")

    # Run
    if throw_exception:
        if os_type == "Windows":
            mock_start_file.side_effect = OSError()
        else:
            mock_subprocess_run.side_effect = OSError()

    default_main_window._on_menu_action_previously_generated_games()

    # Assert
    if throw_exception:
        mock_message_box.return_value.show.assert_called_once()
    else:
        if os_type == "Windows":
            mock_start_file.assert_called_once()
            mock_message_box.return_value.show.assert_not_called()
        else:
            mock_subprocess_run.assert_called_once()
            mock_message_box.return_value.show.assert_not_called()



import discord.ext.commands
from discord_slash import SlashCommand

import randovania


class RandovaniaBot(discord.ext.commands.Bot):
    def __init__(self, configuration: dict):
        super().__init__("%unused-prefix%")
        self.slash = SlashCommand(self)
        self.configuration = configuration
        self.load_extension("randovania.server.discord.preset_lookup")
        self.load_extension("randovania.server.discord.database_command")


def run():
    configuration = randovania.get_configuration()["discord_bot"]

    client = RandovaniaBot(configuration)
    client.run(configuration["token"])



import os
import pprint
import shutil
import typing
from pathlib import Path
from random import Random
from typing import List, Optional

import SuperDuperMetroid.ROM_Patcher
import SuperDuperMetroid.SM_Constants

from randovania.game_description import default_database
from randovania.game_description.assignment import PickupTarget
from randovania.games.game import RandovaniaGame
from randovania.games.patcher import Patcher
from randovania.games.prime.patcher_file_lib import pickup_exporter
from randovania.generator.item_pool import pickup_creator
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.super_metroid.super_metroid_configuration import SuperMetroidConfiguration
from randovania.layout.super_metroid.super_metroid_cosmetic_patches import SuperMetroidCosmeticPatches
from randovania.lib.status_update_lib import ProgressUpdateCallable


_multiplier_for_item = {
    "Energy Tank": 100, "Reserve Tank": 100,
}


def sm_pickup_details_to_patcher(detail: pickup_exporter.ExportedPickupDetails,
                                 ) -> dict:
    if detail.model.game == RandovaniaGame.SUPER_METROID:
        model_name = detail.model.name
    else:
        model_name = "Nothing"

    scan_text = detail.scan_text
    hud_text = detail.hud_text[0]
    pickup_type = "Nothing"
    count = 0

    for resource, quantity in detail.conditional_resources[0].resources:
        if resource.index >= 1000:
            continue
        pickup_type = resource.long_name
        count = quantity
        break

    count *= _multiplier_for_item.get(pickup_type, 1)

    _mapping = {
        "Missile": "Missile Expansion",
        "Super Missile": "Super Missile Expansion",
        "Power Bombs": "Power Bomb Expansion",

        "Nothing": "Missile Expansion",
    }

    result = {
        "item_name": _mapping.get(pickup_type, pickup_type),
        "quantity": count,
        "pickup_index": detail.index.index,
    }

    return result


def _json_to_pickup_placement_data(pickup: dict) -> SuperDuperMetroid.ROM_Patcher.PickupPlacementData:
    return SuperDuperMetroid.ROM_Patcher.PickupPlacementData(
        quantity_given=pickup["quantity"],
        pickup_index=pickup["pickup_index"],
        item_name=pickup["item_name"],
    )


class SuperDuperMetroidPatcher(Patcher):
    @property
    def is_busy(self) -> bool:
        """
        Checks if the patcher is busy right now
        """
        return False

    @property
    def export_can_be_aborted(self) -> bool:
        """
        Checks if patch_game can be aborted
        """
        return False

    @property
    def uses_input_file_directly(self) -> bool:
        """
        Does this patcher uses the input file directly?
        """
        return False

    def has_internal_copy(self, game_files_path: Path) -> bool:
        """
        Checks if the internal storage has an usable copy of the game
        """
        return game_files_path.joinpath("super_metroid", "vanilla.smc").is_file()

    def delete_internal_copy(self, game_files_path: Path):
        """
        Deletes any copy of the game in the internal storage.
        """
        game_files_path.joinpath("super_metroid", "vanilla.smc").unlink(missing_ok=True)

    def default_output_file(self, seed_hash: str) -> str:
        """
        Provides a output file name with the given seed hash.
        :param seed_hash:
        :return:
        """
        return "SM Randomizer - {}".format(seed_hash)

    @property
    def valid_input_file_types(self) -> List[str]:
        return ["smc", "sfc"]

    @property
    def valid_output_file_types(self) -> List[str]:
        return ["smc", "sfc"]

    def create_patch_data(self, description: LayoutDescription, players_config: PlayersConfiguration,
                          cosmetic_patches: SuperMetroidCosmeticPatches) -> dict:
        """
        Creates a JSON serializable dict that can be used to patch the game.
        Intended to be ran on the server for multiworld.
        :return:
        """
        patches = description.all_patches[players_config.player_index]
        db = default_database.game_description_for(RandovaniaGame.SUPER_METROID)
        preset = description.permalink.get_preset(players_config.player_index)
        configuration = typing.cast(SuperMetroidConfiguration, preset.configuration)
        rng = Random(description.permalink.seed_number)

        useless_target = PickupTarget(pickup_creator.create_nothing_pickup(db.resource_database),
                                      players_config.player_index)

        pickup_list = pickup_exporter.export_all_indices(
            patches,
            useless_target,
            db.world_list,
            rng,
            configuration.pickup_model_style,
            configuration.pickup_model_data_source,
            exporter=pickup_exporter.create_pickup_exporter(db, pickup_exporter.GenericAcquiredMemo(), players_config),
            visual_etm=pickup_creator.create_visual_etm(),
        )

        return {
            "item_list": [
                sm_pickup_details_to_patcher(detail)
                for detail in pickup_list
            ],
            "starting_items": {
                item.long_name: quantity * _multiplier_for_item.get(item.long_name, 1)
                for item, quantity in patches.starting_items.items()
            },
        }

    def patch_game(self, input_file: Optional[Path], output_file: Path, patch_data: dict,
                   internal_copies_path: Path, progress_update: ProgressUpdateCallable):
        """
        Invokes the necessary tools to patch the game.
        :param input_file: Vanilla copy of the game. Required if uses_input_file_directly or has_internal_copy is False.
        :param output_file: Where a modified copy of the game is placed.
        :param patch_data: Data created by create_patch_data.
        :param internal_copies_path: Path to where all internal copies are stored.
        :param progress_update: Pushes updates as slow operations are done.
        :return: None
        """
        internal_copy = internal_copies_path.joinpath("super_metroid", "vanilla.smc")
        temporary_output = internal_copies_path.joinpath("super_metroid", "modified.smc")

        if input_file is not None:
            vanilla_bytes = input_file.read_bytes()
            if len(vanilla_bytes) == 0x300200:
                vanilla_bytes = vanilla_bytes[0x200:]

            if len(vanilla_bytes) != 0x300000:
                raise ValueError("Invalid input ROM")

            internal_copy.parent.mkdir(parents=True, exist_ok=True)
            internal_copy.write_bytes(vanilla_bytes)
        else:
            if not internal_copy.is_file():
                raise ValueError("Missing input ROM")
            vanilla_bytes = internal_copy.read_bytes()

        temporary_output.write_bytes(vanilla_bytes)

        item_list = [
            _json_to_pickup_placement_data(pickup)
            for pickup in patch_data["item_list"]
        ]
        starting_items = [
            SuperDuperMetroid.ROM_Patcher.PickupPlacementData(
                quantity_given=quantity,
                pickup_index=-1,
                item_name=name,
            )
            for name, quantity in patch_data["starting_items"]
        ]

        patches_to_apply = ["InstantG4", "MaxAmmoDisplay"]
        SuperDuperMetroid.ROM_Patcher.patch_rom(
            os.fspath(temporary_output),
            item_list=item_list,
            starting_items=starting_items,
            skip_intro=True,
            static_patches=patches_to_apply,
        )

        shutil.copy2(temporary_output, output_file)



from randovania.game_description.world.area_location import AreaLocation

# These are hard-coded into the Randomprime API
# Used by the prime1 patcher
RANDOM_PRIME_CUSTOM_NAMES = {
    AreaLocation(3241871825, 2472970646): 'Crater Entry Point',  # 2818083
    AreaLocation(2831049361, 3222157185): 'Phendrana Drifts North\0(Phendrana Shorelines)',  # 45
    AreaLocation(2831049361, 3708487481): 'Phendrana Drifts South\0(Quarantine Cave)',  # 1900634
    AreaLocation(361692695, 3508802073): 'Exterior Docking Hangar',  # 134218592
    AreaLocation(1056449404, 1005235657): 'Magmoor Caverns North\0(Lava Lake)',  # 31
    AreaLocation(1056449404, 3702104715): 'Magmoor Caverns West\0(Monitor Station)',  # 852002
    AreaLocation(1056449404, 1279075404): 'Magmoor Caverns East\0(Twin Fires)',  # 1048608
    AreaLocation(1056449404, 4012840000): 'Magmoor Caverns South\0(Magmoor Workstation, Debris)',  # 1703972
    AreaLocation(1056449404, 3249312307): 'Magmoor Caverns South\0(Magmoor Workstation, Save Station)',  # 1769512
    AreaLocation(2980859237, 1125030300): 'Phazon Mines East\0(Main Quarry)',  # 28
    AreaLocation(2980859237, 3804417848): 'Phazon Mines West\0(Phazon Processing Center)',  # 1638417
    AreaLocation(972217896, 295707720): 'Tallon Overworld North\0(Tallon Canyon)',  # 917509
    AreaLocation(972217896, 597223686): 'Artifact Temple',  # 1049306
    AreaLocation(972217896, 2318493278): 'Tallon Overworld East\0(Frigate Crash Site)',  # 1441848
    AreaLocation(972217896, 366411659): 'Tallon Overworld West\0(Root Cave)',  # 1507378
    AreaLocation(972217896, 212145392): 'Tallon Overworld South\0(Great Tree Hall, Upper)',  # 2687012
    AreaLocation(972217896, 2098226800): 'Tallon Overworld South\0(Great Tree Hall, Lower)',  # 2818083
    AreaLocation(2214002543, 1047210935): 'Chozo Ruins West\0(Main Plaza)',  # 125
    AreaLocation(2214002543, 2199318005): 'Chozo Ruins North\0(Sun Tower)',  # 1572903
    AreaLocation(2214002543, 2784651681): 'Chozo Ruins East\0(Reflecting Pool, Save Station)',  # 4063276
    AreaLocation(2214002543, 594418447): 'Chozo Ruins South\0(Reflecting Pool, Far End)',  # 4128808

    AreaLocation(0x13d79165, 0xb4b41c48): 'Credits',
}

# Names for use when "elevator" isn't implied
UI_CUSTOM_NAMES = {
    # Ruins
    1047210935: "Main Plaza Elevator",
    2199318005: "Sun Tower Elevator",
    2784651681: "Reflecting Pool West Elevator (Save)",
    594418447: "Reflecting Pool South Elevator (Far)",

    # Magmoor
    1005235657: "Lava Lake Elevator",
    3702104715: "Monitor Station Elevator",
    1279075404: "Twin Fires Elevator",
    4012840000: "Magmoor Workstation West Elevator (Debris)",
    3249312307: "Magmoor Workstation South Elevator (Save)",

    # Mines
    1125030300: "Phazon Processing Elevator",
    3804417848: "Main Quarry Elevator",
    
    # Phendrana
    3222157185: "Phendrana Shoreline Elevator",
    3708487481: "Quarantine Cave Elevator",

    # Tallon
    295707720: "Tallon Canyon Elevator",
    2318493278: "Frigate Crash Site Elevator",
    366411659: "Root Cave Elevator",
    212145392: "Lower/North Great Tree Elevator",
    2098226800: "Upper/South Great Tree Elevator"
}

# Names for use when "elevator" is implied
SHORT_UI_CUSTOM_NAMES = {
    # Ruins
    1047210935: "Main Plaza",
    2199318005: "Sun Tower",
    2784651681: "Reflecting Pool West (Save)",
    594418447: "Reflecting Pool South (Far)",

    # Magmoor
    1005235657: "Lava Lake",
    3702104715: "Monitor Station",
    1279075404: "Twin Fires",
    4012840000: "Workstation West (Debris)",
    3249312307: "Workstation South (Save)",

    # Mines
    1125030300: "Phazon Processing",
    3804417848: "Main Quarry",
    
    # Phendrana
    3222157185: "Phendrana Shoreline",
    3708487481: "Quarantine Cave",

    # Tallon
    295707720: "Tallon Canyon",
    2318493278: "Frigate Crash Site",
    366411659: "Root Cave",
    212145392: "Lower/North Great Tree",
    2098226800: "Upper/South Great Tree"
}


import asyncio
import multiprocessing
from concurrent.futures.process import ProcessPoolExecutor
from multiprocessing.connection import Connection
from typing import Callable, Optional

from randovania.generator import generator
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink
from randovania.resolver import debug


def _generate_layout_worker(output_pipe: Connection,
                            debug_level: int,
                            extra_args: dict):
    def status_update(message: str):
        output_pipe.send(message)
        if output_pipe.poll():
            raise RuntimeError(output_pipe.recv())

    debug.set_level(debug_level)
    return asyncio.run(generator.generate_and_validate_description(status_update=status_update, **extra_args))


def generate_description(permalink: Permalink,
                         status_update: Callable[[str], None],
                         validate_after_generation: bool,
                         timeout_during_generation: bool,
                         attempts: Optional[int],
                         ) -> LayoutDescription:
    receiving_pipe, output_pipe = multiprocessing.Pipe(True)

    debug_level = debug.debug_level()
    if not permalink.spoiler:
        debug_level = 0

    def on_done(_):
        output_pipe.send(None)

    with ProcessPoolExecutor(max_workers=1) as executor:
        extra_args = {
            "permalink": permalink,
            "validate_after_generation": validate_after_generation,
        }
        if not timeout_during_generation:
            extra_args["timeout"] = None
        if attempts is not None:
            extra_args["attempts"] = attempts

        future = executor.submit(_generate_layout_worker, output_pipe, debug_level, extra_args)
        future.add_done_callback(on_done)

        while not future.done():
            message = receiving_pipe.recv()
            if message is not None:
                try:
                    status_update(message)
                except Exception:
                    receiving_pipe.send("close")
                    raise

        return future.result()



import json
from pathlib import Path
from typing import Union
from unittest.mock import patch, MagicMock, call, ANY

import pytest

from randovania.games.patchers import claris_randomizer
from randovania.games.patchers.exceptions import ExportFailure
from randovania.layout.layout_description import LayoutDescription

LayoutDescriptionMock = Union[MagicMock, LayoutDescription]


class CustomException(Exception):
    @classmethod
    def do_raise(cls, x):
        raise CustomException("test exception")


@pytest.fixture(name="valid_tmp_game_root")
def _valid_tmp_game_root(tmp_path):
    game_root = tmp_path.joinpath("game_root")
    game_root.joinpath("files").mkdir(parents=True)
    game_root.joinpath("sys").mkdir()

    for f in ['default.dol', 'FrontEnd.pak', 'Metroid1.pak', 'Metroid2.pak']:
        game_root.joinpath("files", f).write_bytes(b"")

    game_root.joinpath("sys", 'main.dol').write_bytes(b"")

    return game_root


@patch("randovania.games.patchers.csharp_subprocess.process_command", autospec=True)
def test_run_with_args_success(mock_process_command: MagicMock,
                               ):
    # Setup
    args = [MagicMock(), MagicMock()]
    finish_string = "We are done!"
    status_update = MagicMock()
    lines = [
        "line 1",
        "line 2",
        finish_string,
        "post line"
    ]

    def side_effect(_, __, read_callback):
        for line in lines:
            read_callback(line)

    mock_process_command.side_effect = side_effect

    # Run
    claris_randomizer._run_with_args(args, "", finish_string, status_update)

    # Assert
    mock_process_command.assert_called_once_with([str(x) for x in args], "", ANY)
    status_update.assert_has_calls([
        call("line 1"),
        call("line 2"),
        call(finish_string),
    ])


@patch("randovania.games.patchers.csharp_subprocess.process_command", autospec=True)
def test_run_with_args_failure(mock_process_command: MagicMock):
    # Setup
    input_data = "asdf"
    finish_string = "We are done!"
    status_update = MagicMock()
    lines = [
        "line 1",
        "line 2",
        "post line"
    ]

    def side_effect(_, __, read_callback):
        for line in lines:
            read_callback(line)

    mock_process_command.side_effect = side_effect

    # Run
    with pytest.raises(ExportFailure) as error:
        claris_randomizer._run_with_args([], input_data, finish_string, status_update)

    # Assert
    mock_process_command.assert_called_once_with([], input_data, ANY)
    status_update.assert_has_calls([
        call("line 1"),
        call("line 2"),
        call("post line"),
    ])
    assert str(error.value) == "External tool did not send '{}'.".format(finish_string)


@patch("randovania.games.patchers.claris_randomizer.validate_game_files_path", autospec=True)
@patch("randovania.games.patchers.claris_randomizer.get_data_path", autospec=True)
def test_base_args(mock_get_data_path: MagicMock,
                   mock_validate_game_files_path: MagicMock,
                   ):
    # Setup
    mock_get_data_path.return_value = Path("data")
    game_root = Path("root")

    # Run
    results = claris_randomizer._base_args(game_root)

    # Assert
    expected_results = [
        Path("data", "ClarisPrimeRandomizer", "Randomizer.exe"),
        Path("root"),
        "-data:" + str(Path("data", "ClarisPrimeRandomizer", "CustomRandomizerData.json"))
    ]

    assert results == expected_results
    mock_validate_game_files_path.assert_called_once_with(Path("root", "files"))


@pytest.mark.parametrize("has_menu_mod", [False, True])
@patch("shutil.copy", autospec=True)
def test_ensure_no_menu_mod(mock_copy: MagicMock,
                            tmpdir,
                            has_menu_mod: bool,
                            ):
    # Setup
    game_root = Path(tmpdir.join("root"))
    backup_files_path = Path(tmpdir.join("backup"))
    status_update = MagicMock()
    files_folder = game_root.joinpath("files")
    mod_txt = files_folder.joinpath("menu_mod.txt")
    paks = claris_randomizer._ECHOES_PAKS

    if has_menu_mod:
        mod_txt.parent.mkdir(parents=True)
        mod_txt.write_bytes(b"")

    pak_folder = backup_files_path.joinpath("paks")
    pak_folder.mkdir(parents=True)
    for pak in paks:
        pak_folder.joinpath(pak).write_bytes(b"")

    # Run
    claris_randomizer.restore_pak_backups(game_root, backup_files_path, status_update)

    # Assert
    assert not mod_txt.exists()

    status_update.assert_has_calls([
        call("Restoring {} from backup".format(pak_name), i / len(paks))
        for i, pak_name in enumerate(paks)
    ])
    mock_copy.assert_has_calls([
        call(backup_files_path.joinpath("paks", pak_name), files_folder.joinpath(pak_name))
        for pak_name in paks
    ])


@pytest.mark.parametrize("existing", [False, True])
@patch("shutil.copy", autospec=True)
def test_create_pak_backups(mock_copy: MagicMock,
                            tmpdir,
                            existing: bool
                            ):
    # Setup
    game_root = Path(tmpdir.join("root"))
    backup_files_path = Path(tmpdir.join("backup"))
    status_update = MagicMock()
    pak_folder = backup_files_path.joinpath("paks")

    if existing:
        pak_folder.mkdir(parents=True)
        for pak in claris_randomizer._ECHOES_PAKS:
            pak_folder.joinpath(pak).write_bytes(b"")

    # Run
    claris_randomizer.create_pak_backups(game_root, backup_files_path, status_update)

    # Assert
    status_update.assert_has_calls([
        call("Backing up {}".format(pak), i / len(claris_randomizer._ECHOES_PAKS))
        for i, pak in enumerate(claris_randomizer._ECHOES_PAKS)
    ])
    mock_copy.assert_has_calls([
        call(game_root.joinpath("files", pak), pak_folder.joinpath(pak))
        for pak in claris_randomizer._ECHOES_PAKS
    ])


@patch("randovania.games.patchers.claris_randomizer._run_with_args", autospec=True)
@patch("randovania.games.patchers.claris_randomizer.get_data_path", autospec=True)
def test_add_menu_mod_to_files(mock_get_data_path: MagicMock,
                               mock_run_with_args: MagicMock,
                               tmpdir,
                               ):
    # Setup
    mock_get_data_path.return_value = Path("data")
    game_root = Path(tmpdir.join("root"))
    status_update = MagicMock()
    game_root.joinpath("files").mkdir(parents=True)

    # Run
    claris_randomizer._add_menu_mod_to_files(game_root, status_update)

    # Assert
    mock_run_with_args.assert_called_once_with(
        [Path("data", "ClarisEchoesMenu", "EchoesMenu.exe"), game_root.joinpath("files")],
        "",
        "Done!",
        status_update
    )
    assert game_root.joinpath("files", "menu_mod.txt").is_file()


@pytest.mark.parametrize("include_menu_mod", [False, True])
def test_apply_patcher_file(
        include_menu_mod: bool,
        valid_tmp_game_root,
        mocker,
):
    # Setup
    mock_add_menu_mod_to_files: MagicMock = mocker.patch(
        "randovania.games.patchers.claris_randomizer._add_menu_mod_to_files", autospec=True
    )
    mock_run_with_args: MagicMock = mocker.patch(
        "randovania.games.patchers.claris_randomizer._run_with_args", autospec=True
    )
    mock_apply_patches: MagicMock = mocker.patch(
        "randovania.games.prime.echoes_dol_patcher.apply_patches", autospec=True
    )
    mock_create_progress_update_from_successive_messages: MagicMock = mocker.patch(
        "randovania.lib.status_update_lib.create_progress_update_from_successive_messages", autospec=True
    )

    game_root = valid_tmp_game_root
    progress_update = MagicMock()
    status_update = mock_create_progress_update_from_successive_messages.return_value
    mock_data_from_json = mocker.patch("randovania.games.prime.echoes_dol_patcher.EchoesDolPatchesData.from_json")

    patcher_data = {
        "menu_mod": include_menu_mod,
        "dol_patches": {"key": "special"}
    }
    randomizer_data = {"custom": "data"}
    assert claris_randomizer.get_patch_version(game_root) == 0

    # Run
    claris_randomizer.apply_patcher_file(game_root, patcher_data, randomizer_data, progress_update)

    # Assert
    mock_create_progress_update_from_successive_messages.assert_called_once_with(
        progress_update,
        200 if include_menu_mod else 100
    )
    mock_run_with_args.assert_called_once_with(claris_randomizer._base_args(game_root),
                                               json.dumps(patcher_data), "Randomized!", status_update)
    mock_data_from_json.assert_called_once_with({"key": "special"})
    mock_apply_patches.assert_called_once_with(game_root, mock_data_from_json.return_value)

    if include_menu_mod:
        mock_add_menu_mod_to_files.assert_called_once_with(game_root, status_update)
    else:
        mock_add_menu_mod_to_files.assert_not_called()

    assert claris_randomizer.get_patch_version(game_root) == claris_randomizer.CURRENT_PATCH_VERSION
    assert claris_randomizer._get_custom_data_path()



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from PySide2 import QtCore

from randovania.gui.dialog.game_input_dialog import GameInputDialog
from randovania.games.game import RandovaniaGame
from randovania.interface_common.options import Options


@pytest.mark.parametrize("has_output_dir", [False, True])
def test_on_output_file_button_exists(skip_qtbot, tmp_path, mocker, has_output_dir):
    # Setup
    mock_prompt = mocker.patch("randovania.gui.lib.common_qt_lib.prompt_user_for_output_file", autospec=True)

    if has_output_dir:
        output_directory = tmp_path.joinpath("output_path")
        expected_default_name = str(tmp_path.joinpath("output_path", "MyHashGame.iso"))
        output_directory.mkdir()
    else:
        output_directory = None
        expected_default_name = "MyHashGame.iso"

    patcher = MagicMock()
    patcher.default_output_file.return_value = "MyHashGame.iso"
    options = MagicMock()
    options.options_for_game.return_value.output_directory = output_directory
    options.options_for_game.return_value.output_format = "iso"
    window = GameInputDialog(options, patcher, "MyHash", True, RandovaniaGame.METROID_PRIME_ECHOES)
    mock_prompt.return_value = tmp_path.joinpath("foo", "game.iso")

    # Run
    skip_qtbot.mouseClick(window.output_file_button, QtCore.Qt.LeftButton)

    # Assert
    patcher.default_output_file.assert_called_once_with("MyHash")
    mock_prompt.assert_called_once_with(window, expected_default_name + ".iso",
                                        patcher.valid_output_file_types)
    assert window.output_file_edit.text() == str(tmp_path.joinpath("foo", "game.iso"))
    assert tmp_path.joinpath("foo").is_dir()


def test_on_output_file_button_cancel(skip_qtbot, tmpdir, mocker):
    # Setup
    mock_prompt = mocker.patch("randovania.gui.lib.common_qt_lib.prompt_user_for_output_file", autospec=True)

    patcher = MagicMock()
    options = MagicMock()
    options.options_for_game.return_value.output_directory = None
    options.options_for_game.return_value.output_format = "iso"
    patcher.default_output_file.return_value = "Echoes Randomizer - MyHash"
    window = GameInputDialog(options, patcher, "MyHash", True, RandovaniaGame.METROID_PRIME_ECHOES)
    mock_prompt.return_value = None

    # Run
    skip_qtbot.mouseClick(window.output_file_button, QtCore.Qt.LeftButton)

    # Assert
    patcher.default_output_file.assert_called_once_with("MyHash")
    mock_prompt.assert_called_once_with(window, patcher.default_output_file.return_value + ".iso",
                                        patcher.valid_output_file_types)
    assert window.output_file_edit.text() == ""


def test_save_options(skip_qtbot, tmp_path):
    options = Options(tmp_path)
    patcher = MagicMock()
    patcher.valid_output_file_types = ["iso"]
    window = GameInputDialog(options, patcher, "MyHash", True, RandovaniaGame.METROID_PRIME_ECHOES)
    window.output_file_edit.setText("somewhere/game.iso")

    # Run
    window.save_options()

    # Assert
    assert options.options_for_game(RandovaniaGame.METROID_PRIME_ECHOES).output_directory == Path("somewhere")



import math
import pprint
import typing
from random import Random
from typing import Tuple, Iterator, Set, AbstractSet, Dict, \
    Mapping, FrozenSet, Callable, List, Optional

from randovania.game_description.assignment import PickupTarget
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import Hint, HintType
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.generator import reach_lib
from randovania.generator.filler.action import Action, action_name
from randovania.generator.filler.filler_library import UnableToGenerate, should_have_hint, UncollectedState, \
    find_node_with_resource
from randovania.generator.filler.filler_logging import debug_print_collect_event
from randovania.generator.filler.player_state import PlayerState
from randovania.generator.generator_reach import GeneratorReach
from randovania.resolver import debug
from randovania.resolver.random_lib import select_element_with_weight

_DANGEROUS_ACTION_MULTIPLIER = 0.75
_EVENTS_WEIGHT_MULTIPLIER = 0.5
_INDICES_WEIGHT_MULTIPLIER = 1
_LOGBOOKS_WEIGHT_MULTIPLIER = 1
_VICTORY_WEIGHT = 1000
WeightedLocations = Dict[Tuple["PlayerState", PickupIndex], float]


def _calculate_reach_for_progression(reach: GeneratorReach,
                                     progressions: Iterator[PickupEntry],
                                     ) -> GeneratorReach:
    return reach_lib.advance_to_with_reach_copy(reach, reach.state.assign_pickups_resources(progressions))


def _calculate_uncollected_index_weights(uncollected_indices: AbstractSet[PickupIndex],
                                         assigned_indices: AbstractSet[PickupIndex],
                                         seen_counts: Mapping[PickupIndex, int],
                                         indices_groups: List[Set[PickupIndex]],
                                         ) -> Dict[PickupIndex, float]:
    result = {}

    for indices in indices_groups:
        weight_from_collected_indices = math.sqrt(len(indices) / ((1 + len(assigned_indices & indices)) ** 2))

        for index in sorted(uncollected_indices & indices):
            weight_from_seen_count = min(10, seen_counts[index]) ** -2
            result[index] = weight_from_collected_indices * weight_from_seen_count
            # print(f"## {index} : {weight_from_collected_indices} ___ {weight_from_seen_count}")

    return result


def _get_next_player(rng: Random, player_states: List[PlayerState], num_indices: int) -> Optional[PlayerState]:
    """
    Gets the next player a pickup should be placed for.
    :param rng:
    :param player_states:
    :param num_indices: The number of indices av
    :return:
    """
    all_uncollected: Dict[PlayerState, UncollectedState] = {
        player_state: UncollectedState.from_reach(player_state.reach)
        for player_state in player_states
    }

    max_actions = max(player_state.num_actions for player_state in player_states)
    max_uncollected = max(len(uncollected.indices) for uncollected in all_uncollected.values())

    def _calculate_weight(player: PlayerState) -> float:
        return 1 + (max_actions - player.num_actions) * (max_uncollected - len(all_uncollected[player].indices))

    weighted_players = {
        player_state: _calculate_weight(player_state)
        for player_state in player_states
        if not player_state.victory_condition_satisfied() and player_state.potential_actions(num_indices)
    }
    if weighted_players:
        if debug.debug_level() > 1:
            print(f">>>>> Player Weights: {weighted_players}")

        return select_element_with_weight(weighted_players, rng)
    else:
        if all(player_state.victory_condition_satisfied() for player_state in player_states):
            debug.debug_print("Finished because we can win")
            return None
        else:
            total_actions = sum(player_state.num_actions for player_state in player_states)
            unfinished_players = ", ".join([str(player_state) for player_state in player_states
                                            if not player_state.victory_condition_satisfied()])

            raise UnableToGenerate(
                f"{unfinished_players} with no possible actions after {total_actions} total actions."
            )


def weighted_potential_actions(player_state: PlayerState, status_update: Callable[[str], None],
                               num_available_indices: int) -> Dict[Action, float]:
    """
    Weights all potential actions based on current criteria.
    :param player_state:
    :param status_update:
    :param num_available_indices: The number of indices available for placement.
    :return:
    """
    actions_weights: Dict[Action, float] = {}
    current_uncollected = UncollectedState.from_reach(player_state.reach)

    actions = player_state.potential_actions(num_available_indices)
    options_considered = 0

    def update_for_option():
        nonlocal options_considered
        options_considered += 1
        status_update("Checked {} of {} options.".format(options_considered, len(actions)))

    for action in actions:
        if isinstance(action, tuple):
            pickups = typing.cast(Tuple[PickupEntry, ...], action)
            base_weight = _calculate_weights_for(_calculate_reach_for_progression(player_state.reach, pickups),
                                                 current_uncollected)

            multiplier = sum(pickup.probability_multiplier for pickup in pickups) / len(pickups)
            offset = sum(pickup.probability_offset for pickup in pickups)
            weight = (base_weight * multiplier + offset) / len(pickups)

        else:
            weight = _calculate_weights_for(
                reach_lib.advance_to_with_reach_copy(player_state.reach, player_state.reach.state.act_on_node(action)),
                current_uncollected) * _DANGEROUS_ACTION_MULTIPLIER

        actions_weights[action] = weight
        update_for_option()

    if debug.debug_level() > 1:
        for action, weight in actions_weights.items():
            print("{} - {}".format(action_name(action), weight))

    return actions_weights


def retcon_playthrough_filler(rng: Random,
                              player_states: List[PlayerState],
                              status_update: Callable[[str], None],
                              ) -> Tuple[Dict[PlayerState, GamePatches], Tuple[str, ...]]:
    """
    Runs the retcon logic.
    :param rng:
    :param player_states:
    :param status_update:
    :return: A GamePatches for each player and a sequence of placed items.
    """
    debug.debug_print("{}\nRetcon filler started with major items:\n{}".format(
        "*" * 100,
        "\n".join(
            "Player {}: {}".format(
                player_state.index,
                pprint.pformat({
                    item.name: player_state.pickups_left.count(item)
                    for item in sorted(set(player_state.pickups_left), key=lambda item: item.name)
                })
            )
            for player_state in player_states
        )
    ))
    last_message = "Starting."

    def action_report(message: str):
        status_update("{} {}".format(last_message, message))

    for player_state in player_states:
        player_state.update_for_new_state()

    actions_log = []

    while True:
        all_locations_weighted = _calculate_all_pickup_indices_weight(player_states)
        current_player = _get_next_player(rng, player_states, len(all_locations_weighted))
        if current_player is None:
            break

        weighted_actions = weighted_potential_actions(current_player, action_report, len(all_locations_weighted))
        try:
            action = select_element_with_weight(weighted_actions, rng=rng)
        except StopIteration:
            # All actions had weight 0. Select one randomly instead.
            # No need to check if potential_actions is empty, _get_next_player only return players with actions
            action = rng.choice(list(weighted_actions.keys()))

        if isinstance(action, tuple):
            new_pickups: List[PickupEntry] = sorted(action)
            rng.shuffle(new_pickups)

            debug.debug_print(f"\n>>> Will place {len(new_pickups)} pickups")
            for new_pickup in new_pickups:
                log_entry = _assign_pickup_somewhere(new_pickup, current_player, player_states, rng,
                                                     all_locations_weighted)
                actions_log.append(log_entry)
                debug.debug_print(f"* {log_entry}")

                # TODO: this item is potentially dangerous and we should remove the invalidated paths
                current_player.pickups_left.remove(new_pickup)
            current_player.num_actions += 1
        else:
            debug_print_collect_event(action, current_player.game)
            # This action is potentially dangerous. Use `act_on` to remove invalid paths
            current_player.reach.act_on(action)

        last_message = "{} actions performed.".format(sum(player.num_actions for player in player_states))
        status_update(last_message)
        current_player.reach = reach_lib.advance_reach_with_possible_unsafe_resources(current_player.reach)
        current_player.update_for_new_state()

    all_patches = {player_state: player_state.reach.state.patches for player_state in player_states}
    return all_patches, tuple(actions_log)


def _assign_pickup_somewhere(action: PickupEntry,
                             current_player: PlayerState,
                             player_states: List[PlayerState],
                             rng: Random,
                             all_locations_weighted: WeightedLocations,
                             ) -> str:
    """
    Assigns a PickupEntry to a free, collected PickupIndex or as a starting item.
    :param action:
    :param current_player:
    :param player_states:
    :param rng:
    :return:
    """
    assert action in current_player.pickups_left

    if all_locations_weighted and (current_player.num_random_starting_items_placed
                                   >= current_player.configuration.minimum_random_starting_items):

        index_owner_state, pickup_index = select_element_with_weight(all_locations_weighted, rng)
        index_owner_state.assign_pickup(pickup_index, PickupTarget(action, current_player.index))
        all_locations_weighted.pop((index_owner_state, pickup_index))

        # Place a hint for the new item
        hint_location = _calculate_hint_location_for_action(
            action,
            UncollectedState.from_reach(index_owner_state.reach),
            pickup_index,
            rng,
            index_owner_state.scan_asset_initial_pickups,
        )
        if hint_location is not None:
            index_owner_state.reach.state.patches = index_owner_state.reach.state.patches.assign_hint(
                hint_location, Hint(HintType.LOCATION, None, pickup_index))

        if pickup_index in index_owner_state.reach.state.collected_pickup_indices:
            current_player.reach.advance_to(current_player.reach.state.assign_pickup_resources(action))
        else:
            # FIXME: isn't that condition always true?
            pass

        spoiler_entry = pickup_placement_spoiler_entry(current_player.index, action, index_owner_state.game,
                                                       pickup_index, hint_location, index_owner_state.index,
                                                       len(player_states) > 1)

    else:
        current_player.num_random_starting_items_placed += 1
        if (current_player.num_random_starting_items_placed
                > current_player.configuration.maximum_random_starting_items):
            raise UnableToGenerate("Attempting to place more extra starting items than the number allowed.")

        spoiler_entry = f"{action.name} as starting item"
        if len(player_states) > 1:
            spoiler_entry += f" for Player {current_player.index + 1}"
        current_player.reach.advance_to(current_player.reach.state.assign_pickup_to_starting_items(action))

    return spoiler_entry


def _calculate_all_pickup_indices_weight(player_states: List[PlayerState]) -> WeightedLocations:
    all_weights = {}

    total_assigned_pickups = sum(player_state.num_assigned_pickups for player_state in player_states)

    # print("================ WEIGHTS! ==================")
    for player_state in player_states:
        delta = (total_assigned_pickups - player_state.num_assigned_pickups)
        player_weight = 1 + delta

        # print(f"** Player {player_state.index} -- {player_weight}")

        pickup_index_weights = _calculate_uncollected_index_weights(
            player_state.all_indices & UncollectedState.from_reach(player_state.reach).indices,
            set(player_state.reach.state.patches.pickup_assignment),
            player_state.pickup_index_seen_count,
            player_state.indices_groups,
        )
        for pickup_index, weight in pickup_index_weights.items():
            all_weights[(player_state, pickup_index)] = weight * player_weight

    # for (player_state, pickup_index), weight in all_weights.items():
    #     print(f"> {player_state.index} - {pickup_index}: {weight}")
    # print("============================================")

    return all_weights


def _calculate_hint_location_for_action(action: PickupEntry,
                                        current_uncollected: UncollectedState,
                                        pickup_index: PickupIndex,
                                        rng: Random,
                                        scan_asset_initial_pickups: Dict[LogbookAsset, FrozenSet[PickupIndex]],
                                        ) -> Optional[LogbookAsset]:
    """
    Calculates where a hint for the given action should be placed.
    :return: A LogbookAsset to use, or None if no hint should be placed.
    """
    if should_have_hint(action.item_category):
        potential_hint_locations = [
            logbook_asset
            for logbook_asset in current_uncollected.logbooks
            if pickup_index not in scan_asset_initial_pickups[logbook_asset]
        ]
        if potential_hint_locations:
            return rng.choice(sorted(potential_hint_locations))
    return None


def _calculate_weights_for(potential_reach: GeneratorReach,
                           current_uncollected: UncollectedState,
                           ) -> float:
    if potential_reach.victory_condition_satisfied():
        return _VICTORY_WEIGHT

    potential_uncollected = UncollectedState.from_reach(potential_reach) - current_uncollected
    return sum((
        _EVENTS_WEIGHT_MULTIPLIER * int(bool(potential_uncollected.events)),
        _INDICES_WEIGHT_MULTIPLIER * int(bool(potential_uncollected.indices)),
        _LOGBOOKS_WEIGHT_MULTIPLIER * int(bool(potential_uncollected.logbooks)),
    ))


def pickup_placement_spoiler_entry(owner_index: int, action: PickupEntry, game: GameDescription,
                                   pickup_index: PickupIndex, hint: Optional[LogbookAsset],
                                   player_index: int, add_indices: bool) -> str:
    world_list = game.world_list
    if hint is not None:
        hint_string = " with hint at {}".format(
            world_list.node_name(find_node_with_resource(hint, world_list.all_nodes),
                                 with_world=True, distinguish_dark_aether=True))
    else:
        hint_string = ""

    pickup_node = find_node_with_resource(pickup_index, world_list.all_nodes)
    return "{4}{0} at {3}{1}{2}".format(
        action.name,
        world_list.node_name(pickup_node, with_world=True, distinguish_dark_aether=True),
        hint_string,
        f"player {player_index + 1}'s " if add_indices else "",
        f"Player {owner_index + 1}'s " if add_indices else "",
    )



from typing import List, Optional, Dict

import dolphin_memory_engine

from randovania.game_connection.memory_executor_choice import MemoryExecutorChoice
from randovania.game_connection.executor.memory_operation import (MemoryOperationException, MemoryOperation,
                                                                  MemoryOperationExecutor)

MEM1_START = 0x80000000
MEM1_END = 0x81800000


def _validate_range(address: int, size: int):
    if address < MEM1_START or address + size > MEM1_END:
        raise MemoryOperationException(
            f"Range {address:x} -> {address + size:x} is outside of the GameCube memory range.")


class DolphinExecutor(MemoryOperationExecutor):
    def __init__(self):
        super().__init__()
        self.dolphin = dolphin_memory_engine

    @property
    def lock_identifier(self) -> Optional[str]:
        return "randovania-dolphin-backend"

    @property
    def backend_choice(self) -> MemoryExecutorChoice:
        return MemoryExecutorChoice.DOLPHIN

    async def connect(self) -> bool:
        if not self.dolphin.is_hooked():
            self.dolphin.hook()

        return self.dolphin.is_hooked()

    async def disconnect(self):
        raise NotImplementedError()

    def _test_still_hooked(self):
        try:
            if len(self.dolphin.read_bytes(0x0, 4)) != 4:
                raise RuntimeError("Dolphin hook didn't read the correct byte count")
        except RuntimeError as e:
            self.logger.warning(f"Test read for Dolphin hook didn't work: {e}")
            self.dolphin.un_hook()

    def is_connected(self) -> bool:
        if self.dolphin.is_hooked():
            self._test_still_hooked()

        return self.dolphin.is_hooked()

    # Game Backend Stuff
    def _memory_operation(self, op: MemoryOperation, pointers: Dict[int, Optional[int]]) -> Optional[bytes]:
        op.validate_byte_sizes()

        address = op.address
        if op.offset is not None:
            if address not in pointers:
                raise MemoryOperationException(f"Invalid op: {address:x} is not in pointers")

            if pointers[address] is None:
                return None
            address = pointers[address] + op.offset

        _validate_range(address, op.byte_count)

        if not self.dolphin.is_hooked():
            raise MemoryOperationException("Lost connection do Dolphin")

        try:
            result = None
            if op.read_byte_count is not None:
                result = self.dolphin.read_bytes(address, op.read_byte_count)

            if op.write_bytes is not None:
                self.dolphin.write_bytes(address, op.write_bytes)
                self.logger.debug(f"Wrote {op.write_bytes.hex()} to {address:x}")

        except RuntimeError as e:
            raise MemoryOperationException(f"Lost connection do Dolphin: {e}")

        return result

    async def perform_memory_operations(self, ops: List[MemoryOperation]) -> Dict[MemoryOperation, bytes]:
        pointers_to_read = set()
        for op in ops:
            if op.offset is not None:
                pointers_to_read.add(op.address)

        pointers = {}
        for pointer in pointers_to_read:
            if not self.dolphin.is_hooked():
                raise MemoryOperationException("Lost connection do Dolphin")

            try:
                pointers[pointer] = self.dolphin.follow_pointers(pointer, [0])
            except RuntimeError:
                pointers[pointer] = None
                self.logger.debug(f"Failed to read a valid pointer from {pointer:x}")
                self._test_still_hooked()

            if not self.dolphin.is_hooked():
                raise MemoryOperationException("Lost connection do Dolphin")

        result = {}
        for op in ops:
            op_result = self._memory_operation(op, pointers)
            if op_result is not None:
                result[op] = op_result
        return result



def apply_migrations():
    import playhouse.migrate
    import randovania
    from randovania.server import database

    randovania.setup_logging('DEBUG', None)

    configuration = randovania.get_configuration()
    database.db.init(configuration["server_config"]['database_path'])
    database.db.connect(reuse_if_open=True)
    migrator = playhouse.migrate.SqliteMigrator(database.db)

    with database.db.atomic():
        playhouse.migrate.migrate(
            migrator.rename_table("gamesession", "game_session"),
            migrator.rename_table("gamesessionmembership", "game_session_membership"),
            migrator.rename_table("gamesessionpreset", "game_session_preset"),
            migrator.rename_table("gamesessionteamaction", "game_session_team_action"),
            migrator.add_column("game_session", "dev_features", database.GameSession.dev_features),
        )



import asyncio
from argparse import ArgumentParser
from pathlib import Path

from randovania.layout import game_to_class


def randomize_command_logic(args):
    return asyncio.run(randomize_command_logic_async(args))


async def randomize_command_logic_async(args):
    from randovania.games.patcher_provider import PatcherProvider
    from randovania.generator import generator
    from randovania.layout.base.cosmetic_patches import BaseCosmeticPatches
    from randovania.interface_common.players_configuration import PlayersConfiguration
    from randovania.layout.layout_description import LayoutDescription
    from randovania.layout.permalink import Permalink
    from randovania.interface_common.options import Options

    def status_update(s):
        if args.verbose:
            print(s)

    if args.permalink is not None:
        permalink = Permalink.from_str(args.permalink)
        layout_description = await generator.generate_and_validate_description(
            permalink=permalink,
            status_update=status_update,
            validate_after_generation=True,
        )
    else:
        layout_description = LayoutDescription.from_file(args.log_file)

    players_config = PlayersConfiguration(args.player_index,
                                          {i: f"Player {i + 1}"
                                           for i in range(layout_description.permalink.player_count)})
    preset = layout_description.permalink.get_preset(players_config.player_index)

    internal_copies_path = Options.with_default_data_dir().internal_copies_path
    patcher_provider = PatcherProvider()
    cosmetic_patches = game_to_class.GAME_TO_COSMETIC[preset.game].default()
    patcher = patcher_provider.patcher_for_game(preset.game)

    patch_data = patcher.create_patch_data(layout_description, players_config, cosmetic_patches)
    patcher.patch_game(args.input_file, args.output_file, patch_data, internal_copies_path, lambda x, _: status_update(x))


def add_randomize_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "randomize",
        help="Randomizes a game files path."
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument("--permalink", type=str, help="The permalink to use")
    group.add_argument("--log-file", type=Path, help="A seed log file to use")

    parser.add_argument("--verbose", action="store_true", help="Prints progress",
                        default=False)
    parser.add_argument("--player-index", type=int, default=0,
                        help="Number of the player to export, for multiworld games. 0-indexed")

    parser.add_argument(
        "--input-file",
        type=Path,
        help="File to use to randomize the game.")
    parser.add_argument(
        "output_file",
        type=Path,
        help="Output file.")
    parser.set_defaults(func=randomize_command_logic)



from unittest.mock import MagicMock

import pytest
from PySide2 import QtGui

from randovania.gui.lib import theme


def test_set_dark_theme_no_change(skip_qtbot):
    theme._current_dark_theme = False
    theme.set_dark_theme(False)


@pytest.mark.parametrize("active", [False, True])
def test_set_dark_theme_change(skip_qtbot, active, mocker):
    mock_load_stylesheet: MagicMock = mocker.patch("qdarktheme.load_stylesheet", return_value="")

    app = MagicMock()
    app.palette.return_value = QtGui.QPalette()

    theme._current_dark_theme = not active
    theme.set_dark_theme(active, app=app)

    mock_load_stylesheet.assert_called_once_with(theme="dark" if active else "light")
    app.setStyleSheet.assert_called_once()
    app.setPalette.assert_called_once()



from typing import Iterator, TypeVar, Dict, Any, Set, NamedTuple

from randovania.game_description.assignment import PickupAssignment
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import ResourceInfo
from randovania.game_description.world.node import Node, PickupNode, ResourceNode
from randovania.generator import reach_lib
from randovania.generator.generator_reach import GeneratorReach


def filter_pickup_nodes(nodes: Iterator[Node]) -> Iterator[PickupNode]:
    for node in nodes:
        if isinstance(node, PickupNode):
            yield node


def filter_unassigned_pickup_nodes(nodes: Iterator[Node],
                                   pickup_assignment: PickupAssignment,
                                   ) -> Iterator[PickupNode]:
    for node in filter_pickup_nodes(nodes):
        if node.pickup_index not in pickup_assignment:
            yield node


class UnableToGenerate(RuntimeError):
    pass


def should_have_hint(item_category: ItemCategory) -> bool:
    return item_category.is_major


X = TypeVar("X")


def _filter_not_in_dict(elements: Iterator[X],
                        dictionary: Dict[X, Any],
                        ) -> Set[X]:
    return set(elements) - set(dictionary.keys())


class UncollectedState(NamedTuple):
    indices: Set[PickupIndex]
    logbooks: Set[LogbookAsset]
    events: Set[ResourceInfo]

    @classmethod
    def from_reach(cls, reach: GeneratorReach) -> "UncollectedState":
        return UncollectedState(
            _filter_not_in_dict(reach.state.collected_pickup_indices, reach.state.patches.pickup_assignment),
            _filter_not_in_dict(reach.state.collected_scan_assets, reach.state.patches.hints),
            set(reach.state.collected_events)
        )

    def __sub__(self, other: "UncollectedState") -> "UncollectedState":
        return UncollectedState(
            self.indices - other.indices,
            self.logbooks - other.logbooks,
            self.events - other.events
        )


def find_node_with_resource(resource: ResourceInfo,
                            haystack: Iterator[Node],
                            ) -> ResourceNode:
    for node in haystack:
        if isinstance(node, ResourceNode) and node.resource() == resource:
            return node



import dataclasses
from typing import List, Dict, Optional, Callable

from randovania.game_description.resources import search
from randovania.game_description.resources.damage_resource_info import DamageReduction
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_info import ResourceInfo, CurrentResources
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.games.game import RandovaniaGame


def default_base_damage_reduction(db: "ResourceDatabase", current_resources: CurrentResources):
    return 1.0


@dataclasses.dataclass(frozen=True)
class ResourceDatabase:
    game_enum: RandovaniaGame
    item: List[ItemResourceInfo]
    event: List[SimpleResourceInfo]
    trick: List[TrickResourceInfo]
    damage: List[SimpleResourceInfo]
    version: List[SimpleResourceInfo]
    misc: List[SimpleResourceInfo]
    requirement_template: Dict[str, "Requirement"]
    damage_reductions: Dict[SimpleResourceInfo, List[DamageReduction]]
    energy_tank_item_index: int
    item_percentage_index: Optional[int]
    multiworld_magic_item_index: Optional[int]
    base_damage_reduction: Callable[["ResourceDatabase", CurrentResources], float] = default_base_damage_reduction

    def get_by_type(self, resource_type: ResourceType) -> List[ResourceInfo]:
        if resource_type == ResourceType.ITEM:
            return self.item
        elif resource_type == ResourceType.EVENT:
            return self.event
        elif resource_type == ResourceType.TRICK:
            return self.trick
        elif resource_type == ResourceType.DAMAGE:
            return self.damage
        elif resource_type == ResourceType.VERSION:
            return self.version
        elif resource_type == ResourceType.MISC:
            return self.misc
        else:
            raise ValueError(
                "Invalid resource_type: {}".format(resource_type))

    def get_by_type_and_index(self, resource_type: ResourceType,
                              index: int) -> ResourceInfo:
        return search.find_resource_info_with_id(self.get_by_type(resource_type), index, resource_type)

    def get_item(self, index: int) -> ItemResourceInfo:
        return self.get_by_type_and_index(ResourceType.ITEM, index)

    def get_item_by_name(self, name: str) -> ItemResourceInfo:
        return search.find_resource_info_with_long_name(self.item, name)

    @property
    def item_percentage(self) -> Optional[ItemResourceInfo]:
        if self.item_percentage_index is not None:
            return self.get_by_type_and_index(ResourceType.ITEM, self.item_percentage_index)

    @property
    def energy_tank(self) -> ItemResourceInfo:
        return self.get_by_type_and_index(ResourceType.ITEM, self.energy_tank_item_index)

    @property
    def multiworld_magic_item(self) -> Optional[ItemResourceInfo]:
        if self.multiworld_magic_item_index is not None:
            return self.get_item(self.multiworld_magic_item_index)

    def get_damage_reduction(self, resource: SimpleResourceInfo, current_resources: CurrentResources):
        multiplier = self.base_damage_reduction(self, current_resources)

        for reduction in self.damage_reductions.get(resource, []):
            if reduction.inventory_item is None or current_resources.get(reduction.inventory_item, 0) > 0:
                multiplier *= reduction.damage_multiplier

        return multiplier



import re
import struct

import pytest
from mock import AsyncMock, MagicMock

from randovania.dol_patching.assembler import BaseInstruction
from randovania.game_connection.connection_base import InventoryItem
from randovania.game_connection.connector.echoes_remote_connector import EchoesRemoteConnector
from randovania.game_connection.connector.prime_remote_connector import DolRemotePatch
from randovania.game_connection.executor.memory_operation import MemoryOperationException, MemoryOperation
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.game import RandovaniaGame
from randovania.games.prime.echoes_dol_patches import EchoesDolVersion


@pytest.fixture(name="version")
def echoes_version():
    from randovania.games.prime import echoes_dol_versions
    return echoes_dol_versions.ALL_VERSIONS[0]


@pytest.fixture(name="connector")
def echoes_remote_connector(version: EchoesDolVersion):
    connector = EchoesRemoteConnector(version)
    return connector


@pytest.mark.asyncio
async def test_is_this_version(connector: EchoesRemoteConnector):
    # Setup
    build_info = b"!#$MetroidBuildInfo!#$Build v1.028 10/18/2004 10:44:32"
    executor = AsyncMock()
    executor.perform_single_memory_operation.return_value = build_info

    # Run
    assert await connector.is_this_version(executor)


@pytest.mark.parametrize(["message_original", "message_encoded", "previous_size"], [
    ("Magoo", b'\x00M\x00a\x00g\x00o\x00o\x00\x00', 0),
    ("Magoo2", b'\x00M\x00a\x00g\x00o\x00o\x002\x00\x00\x00\x00', 0),
    ("Magoo", b'\x00M\x00a\x00g\x00o\x00o\x00 \x00\x00\x00\x00', 10),
])
@pytest.mark.asyncio
async def test_write_string_to_game_buffer(connector: EchoesRemoteConnector, version: EchoesDolVersion,
                                           message_original, message_encoded, previous_size):
    # Setup
    connector._last_message_size = previous_size

    # Run
    result = connector._write_string_to_game_buffer(message_original)

    # Assert
    assert result == MemoryOperation(version.string_display.message_receiver_string_ref,
                                     write_bytes=message_encoded)


@pytest.mark.asyncio
async def test_get_inventory_valid(connector: EchoesRemoteConnector):
    # Setup
    executor = AsyncMock()
    executor.perform_memory_operations.side_effect = lambda ops: {
        op: struct.pack(">II", item.max_capacity, item.max_capacity)
        for op, item in zip(ops, connector.game.resource_database.item)
    }

    # Run
    inventory = await connector.get_inventory(executor)

    # Assert
    assert inventory == {
        item: InventoryItem(item.max_capacity, item.max_capacity)
        for item in connector.game.resource_database.item
    }


@pytest.mark.asyncio
async def test_get_inventory_invalid_capacity(connector: EchoesRemoteConnector):
    # Setup
    custom_inventory = {5: InventoryItem(0, 50)}

    executor = AsyncMock()
    executor.perform_memory_operations.side_effect = lambda ops: {
        op: struct.pack(">II",
                        *custom_inventory.get(item.index, InventoryItem(item.max_capacity, item.max_capacity)))
        for op, item in zip(ops, connector.game.resource_database.item)
    }

    # Run
    msg = "Received InventoryItem(amount=0, capacity=50) for Darkburst, which is an invalid state."
    with pytest.raises(MemoryOperationException, match=re.escape(msg)):
        await connector.get_inventory(executor)


@pytest.mark.asyncio
async def test_get_inventory_invalid_amount(connector: EchoesRemoteConnector):
    # Setup
    custom_inventory = {5: InventoryItem(1, 0)}

    executor = AsyncMock()
    executor.perform_memory_operations.side_effect = lambda ops: {
        op: struct.pack(">II",
                        *custom_inventory.get(item.index, InventoryItem(item.max_capacity, item.max_capacity)))
        for op, item in zip(ops, connector.game.resource_database.item)
    }

    # Run
    msg = "Received InventoryItem(amount=1, capacity=0) for Darkburst, which is an invalid state."
    with pytest.raises(MemoryOperationException, match=re.escape(msg)):
        await connector.get_inventory(executor)


@pytest.mark.parametrize("capacity", [0, 10])
@pytest.mark.asyncio
async def test_known_collected_locations_nothing(connector: EchoesRemoteConnector, capacity: int):
    # Setup
    executor = AsyncMock()
    executor.perform_single_memory_operation.return_value = struct.pack(">II", 0, capacity)

    # Run
    locations, patches = await connector.known_collected_locations(executor)

    # Assert
    assert locations == set()
    assert patches == []


@pytest.mark.parametrize("capacity", [0, 10])
@pytest.mark.asyncio
async def test_known_collected_locations_location(connector: EchoesRemoteConnector, version: EchoesDolVersion,
                                                  mocker, capacity):
    # Setup
    mock_item_patch: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.adjust_item_amount_and_capacity_patch")

    executor = AsyncMock()
    executor.perform_single_memory_operation.return_value = struct.pack(">II", 10, 10 + capacity)

    # Run
    locations, patches = await connector.known_collected_locations(executor)

    # Assert
    mock_item_patch.assert_called_once_with(version.powerup_functions,
                                            RandovaniaGame.METROID_PRIME_ECHOES,
                                            connector.game.resource_database.multiworld_magic_item.index,
                                            -10)

    assert locations == {PickupIndex(9)}
    assert patches == [DolRemotePatch([], mock_item_patch.return_value)]


@pytest.mark.asyncio
async def test_find_missing_remote_pickups_nothing(connector: EchoesRemoteConnector):
    # Setup
    executor = AsyncMock()
    inventory = {connector.game.resource_database.multiworld_magic_item: InventoryItem(0, 0)}

    # Run
    patches, has_message = await connector.find_missing_remote_pickups(executor, inventory, [], False)

    # Assert
    assert patches == []
    assert not has_message


@pytest.mark.asyncio
async def test_find_missing_remote_pickups_pending_location(connector: EchoesRemoteConnector):
    # Setup
    executor = AsyncMock()
    inventory = {connector.game.resource_database.multiworld_magic_item: InventoryItem(5, 15)}

    # Run
    patches, has_message = await connector.find_missing_remote_pickups(executor, inventory, [], False)

    # Assert
    assert patches == []
    assert not has_message


@pytest.mark.parametrize("in_cooldown", [False, True])
@pytest.mark.asyncio
async def test_find_missing_remote_pickups_give_pickup(connector: EchoesRemoteConnector, version: EchoesDolVersion,
                                                       mocker, in_cooldown):
    # Setup
    mock_item_patch: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.increment_item_capacity_patch")
    mock_call_display_hud_patch: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.call_display_hud_patch")

    pickup_patches = MagicMock()
    connector._write_string_to_game_buffer = MagicMock()
    connector._patches_for_pickup = AsyncMock(return_value=([pickup_patches, pickup_patches], "The Message"))

    executor = AsyncMock()
    inventory = {connector.game.resource_database.multiworld_magic_item: InventoryItem(0, 0)}
    permanent_pickups = [
        ("A", MagicMock()),
        ("B", MagicMock()),
    ]

    # Run
    patches, has_message = await connector.find_missing_remote_pickups(executor, inventory, permanent_pickups,
                                                                       in_cooldown)

    # Assert
    if in_cooldown:
        assert patches == []
        assert not has_message
        mock_item_patch.assert_not_called()
        connector._patches_for_pickup.assert_not_called()
        connector._write_string_to_game_buffer.assert_not_called()
        mock_call_display_hud_patch.assert_not_called()
        return

    mock_item_patch.assert_called_once_with(version.powerup_functions,
                                            RandovaniaGame.METROID_PRIME_ECHOES,
                                            connector.game.resource_database.multiworld_magic_item.index)
    connector._patches_for_pickup.assert_awaited_once_with(permanent_pickups[0][0], permanent_pickups[0][1], inventory)
    assert has_message
    assert patches == [
        DolRemotePatch([], pickup_patches),
        DolRemotePatch([], pickup_patches),
        DolRemotePatch([], mock_item_patch.return_value),
        DolRemotePatch(
            [connector._write_string_to_game_buffer.return_value],
            mock_call_display_hud_patch.return_value,
        ),
    ]
    connector._write_string_to_game_buffer.assert_called_once_with("The Message")
    mock_call_display_hud_patch.assert_called_once_with(version.string_display)


@pytest.mark.asyncio
async def test_patches_for_pickup(connector: EchoesRemoteConnector, version: EchoesDolVersion, mocker, generic_item_category):
    # Setup
    mock_item_patch: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.adjust_item_amount_and_capacity_patch")

    db = connector.game.resource_database
    pickup = PickupEntry("Pickup", 0, generic_item_category, generic_item_category, progression=tuple(),
                         extra_resources=(
                             (db.energy_tank, db.energy_tank.max_capacity),
                             (db.item_percentage, 1),
                         ))
    inventory = {
        db.multiworld_magic_item: InventoryItem(0, 0),
        db.energy_tank: InventoryItem(1, 1),
    }

    # Run
    patches, message = await connector._patches_for_pickup("Someone", pickup, inventory)

    # Assert
    mock_item_patch.assert_called_once_with(version.powerup_functions,
                                            RandovaniaGame.METROID_PRIME_ECHOES,
                                            db.energy_tank.index,
                                            db.energy_tank.max_capacity)
    assert patches == [mock_item_patch.return_value]
    assert message == "Received Pickup from Someone."


@pytest.mark.asyncio
async def test_execute_remote_patches(connector: EchoesRemoteConnector, version: EchoesDolVersion, mocker):
    # Setup
    patch_address, patch_bytes = MagicMock(), MagicMock()
    mock_remote_execute: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.create_remote_execution_body",
        return_value=(patch_address, patch_bytes)
    )

    executor = AsyncMock()

    memory_op_a = MemoryOperation(1234, write_bytes=b"1234")
    instructions = [BaseInstruction(), BaseInstruction()]
    patches = [
        DolRemotePatch([memory_op_a], instructions[:1]),
        DolRemotePatch([], instructions[1:]),
    ]
    memory_operations = [
        memory_op_a,
        MemoryOperation(patch_address, write_bytes=patch_bytes),
        MemoryOperation(version.cstate_manager_global + 0x2, write_bytes=b"\x01"),
    ]

    # Run
    await connector.execute_remote_patches(executor, patches)

    # Assert
    mock_remote_execute.assert_called_once_with(version.string_display, instructions)
    executor.perform_memory_operations.assert_awaited_once_with(memory_operations)


@pytest.mark.parametrize("correct_vtable", [False, True])
@pytest.mark.parametrize("has_pending_op", [False, True])
@pytest.mark.parametrize("has_world", [False, True])
@pytest.mark.asyncio
async def test_fetch_game_status(connector: EchoesRemoteConnector, version: EchoesDolVersion,
                                 has_world, has_pending_op, correct_vtable):
    # Setup
    expected_world = connector.game.world_list.worlds[0]

    executor = AsyncMock()
    executor.perform_memory_operations.side_effect = lambda ops: {
        ops[0]: expected_world.world_asset_id.to_bytes(4, "big") if has_world else b"DEAD",
        ops[1]: b"\x01" if has_pending_op else b"\x00",
        ops[2]: version.cplayer_vtable.to_bytes(4, "big") if correct_vtable else b"CAFE",
    }

    # Run
    actual_has_op, actual_world = await connector.current_game_status(executor)

    # Assert
    if has_world and correct_vtable:
        assert actual_world is expected_world
    else:
        assert actual_world is None
    assert actual_has_op == has_pending_op



from enum import Enum
from typing import Dict, Optional

from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world.node import PickupNode
from randovania.game_description.world.world_list import WorldList
from randovania.interface_common.players_configuration import PlayersConfiguration


class TextColor(Enum):
    GUARDIAN = "#FF3333"
    ITEM = "#FF6705B3"
    PRIME1_ITEM = "#c300ff"
    JOKE = "#45F731"
    LOCATION = "#FF3333"
    PRIME1_LOCATION = "#89a1ff"
    PLAYER = "#d4cc33"


def color_text(color: TextColor, text: str):
    return f"&push;&main-color={color.value};{text}&pop;"


def create_simple_logbook_hint(asset_id: int, hint: str) -> dict:
    return {
        "asset_id": asset_id,
        "strings": [hint, "", hint],
    }


class AreaNamer:
    world_list: WorldList
    index_to_node: Dict[PickupIndex, PickupNode]

    def __init__(self, world_list: WorldList):
        self.world_list = world_list

        self.index_to_node = {
            node.pickup_index: node
            for node in world_list.all_nodes
            if isinstance(node, PickupNode)
        }

    def location_name(self, pickup_index: PickupIndex, hide_area: bool,
                      color: Optional[TextColor] = TextColor.LOCATION) -> str:
        result = self.node_name(self.index_to_node[pickup_index], hide_area)
        if color is not None:
            return color_text(color, result)
        return result

    def node_name(self, pickup_node: PickupNode, hide_area: bool) -> str:
        if hide_area:
            return self.world_list.world_name_from_node(pickup_node, True)
        else:
            return self.world_list.area_name(self.world_list.nodes_to_area(pickup_node))


class Determiner:
    s: str
    supports_title: bool

    def __init__(self, s, supports_title: bool = True):
        self.s = s
        self.supports_title = supports_title

    def __format__(self, format_spec):
        return self.s.__format__(format_spec)

    @property
    def title(self):
        if self.supports_title:
            return self.s.title()
        else:
            return self.s


def player_determiner(players_config: PlayersConfiguration, player: int,
                      color: TextColor = TextColor.PLAYER) -> Determiner:
    return Determiner(
        "{}'s ".format(color_text(color, players_config.player_names[player])),
        False
    )



from dataclasses import dataclass
from typing import Optional, Iterator, Tuple

from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.resource_info import ResourceGainTuple, ResourceGain, ResourceQuantity, \
    CurrentResources
from randovania.games.game import RandovaniaGame


@dataclass(frozen=True)
class ResourceConversion:
    source: ItemResourceInfo
    target: ItemResourceInfo
    clear_source: bool = True
    overwrite_target: bool = False


@dataclass(frozen=True)
class ResourceLock:
    locked_by: ItemResourceInfo
    item_to_lock: ItemResourceInfo
    temporary_item: ItemResourceInfo

    def convert_gain(self, gain: ResourceGain) -> ResourceGain:
        for resource, quantity in gain:
            if self.item_to_lock == resource:
                resource = self.temporary_item
            yield resource, quantity

    def unlock_conversion(self) -> ResourceConversion:
        return ResourceConversion(source=self.temporary_item,
                                  target=self.item_to_lock)


@dataclass(frozen=True)
class ConditionalResources:
    name: Optional[str]
    item: Optional[ItemResourceInfo]
    resources: ResourceGainTuple


@dataclass(frozen=True)
class PickupModel:
    game: RandovaniaGame
    name: str


@dataclass(frozen=True)
class PickupEntry:
    name: str
    model: PickupModel
    item_category: ItemCategory
    broad_category: ItemCategory
    progression: ResourceGainTuple
    extra_resources: ResourceGainTuple = tuple()
    unlocks_resource: bool = False
    resource_lock: Optional[ResourceLock] = None
    respects_lock: bool = True
    probability_offset: float = 0
    probability_multiplier: float = 1
    required_progression: int = 0

    def __post_init__(self):
        if not isinstance(self.progression, tuple):
            raise ValueError("resources should be a tuple, got {}".format(self.progression))

        for i, progression in enumerate(self.progression):
            if not isinstance(progression, tuple):
                raise ValueError(f"{i}-th progression should be a tuple, got {progression}")

            if len(progression) != 2:
                raise ValueError(f"{i}-th progression should have 2 elements, got {len(progression)}")

            if not isinstance(progression[1], int):
                raise ValueError(f"{i}-th progression second field should be a int, got {progression[1]}")

            if progression[1] > progression[0].max_capacity:
                raise ValueError(f"{i}-th progression has {progression[1]} quantity, "
                                 f"higher than max for {progression[0]}")

        for resource, quantity in self.extra_resources:
            if isinstance(resource, ItemResourceInfo):
                if quantity > resource.max_capacity:
                    raise ValueError(f"Attempt to give {quantity} of {resource.long_name}, more than max capacity")

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        return self.name < other.name

    @property
    def conditional_resources(self):
        previous: Optional[ItemResourceInfo] = None
        for progression in self.progression:
            yield ConditionalResources(
                name=progression[0].long_name,
                item=previous,
                resources=(progression,) + self.extra_resources,
            )
            previous = progression[0]

        if not self.progression:
            yield ConditionalResources(
                name=self.name,
                item=None,
                resources=self.extra_resources,
            )

    @property
    def convert_resources(self) -> Tuple[ResourceConversion, ...]:
        if self.unlocks_resource and self.resource_lock is not None:
            return (self.resource_lock.unlock_conversion(),)
        else:
            return tuple()

    def conditional_for_resources(self, current_resources: CurrentResources) -> ConditionalResources:
        last_conditional: Optional[ConditionalResources] = None

        for conditional in self.conditional_resources:
            if conditional.item is None or current_resources.get(conditional.item, 0) > 0:
                last_conditional = conditional
            else:
                break

        assert last_conditional is not None
        return last_conditional

    def conversion_resource_gain(self, current_resources: CurrentResources) -> ResourceGain:
        for conversion in self.convert_resources:
            quantity = current_resources.get(conversion.source, 0)
            yield conversion.source, -quantity
            yield conversion.target, quantity

    def resource_gain(self, current_resources: CurrentResources, force_lock: bool = False) -> ResourceGain:
        resources = self.conditional_for_resources(current_resources).resources

        if (force_lock or self.respects_lock) and not self.unlocks_resource and (
                self.resource_lock is not None and current_resources.get(self.resource_lock.locked_by, 0) == 0):
            yield from self.resource_lock.convert_gain(resources)
        else:
            yield from resources

        yield from self.conversion_resource_gain(current_resources)

    def __str__(self):
        return "Pickup {}".format(self.name)

    @property
    def all_resources(self) -> Iterator[ResourceQuantity]:
        yield from self.progression
        yield from self.extra_resources

    @property
    def is_expansion(self) -> bool:
        return self.item_category.is_expansion



import dataclasses

from randovania.generator.item_pool import pool_creator
from randovania.layout.prime2.echoes_configuration import LayoutSkyTempleKeyMode


def test_calculate_pool_item_count(default_layout_configuration):
    layout_configuration = dataclasses.replace(
        default_layout_configuration,
        major_items_configuration=dataclasses.replace(
            default_layout_configuration.major_items_configuration,
            minimum_random_starting_items=2,
        ),
        sky_temple_keys=LayoutSkyTempleKeyMode.ALL_BOSSES,
    )

    # Run
    result = pool_creator.calculate_pool_item_count(layout_configuration)

    # Assert
    assert result == (118, 121)
